
// Ranges and maps and subscripts and indexes.

#include "RangeRep.h"
#include "CollapsedRangeRep.h"
#include "DimRangeRep.h"
#include "BlockRangeRep.h"
#include "CyclicRangeRep.h"
#include "StepRangeRep.h"
#include "IrregRangeRep.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Range': handle class for range representations.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#define DEF_LB  0
#define DEF_UB  -2

struct Trip {
  Trip() : lb(DEF_LB), ub(DEF_UB), str(1) {}
  Trip(const int _lb) : lb(_lb), ub(DEF_UB), str(1) {}
  Trip(const int _lb, const int _ub) : lb(_lb), ub(_ub), str(1) {}
  Trip(const int _lb, const int _ub, const int _str) :
      lb(_lb), ub(_ub), str(_str) {}

      // probably should check _ub >= -1.

  const int lb, ub, str ;
} ;

class Range {
public :
  // Constructors and destructors

  Range() : rep(0) {}

  Range(const Range& x) : rep(x.rep) {
    if(rep) rep->incr() ;
  }

  ~Range() {if(rep) rep->decr() ; }

  // Subranges

  Range subrng(const int extent,
               const int base, const int stride = 1) const {
    if(extent != 0) {
      const int parExtent = size() ;
      const int top = base + stride * (extent - 1) ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: subrng: subrange out of range") ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  Range triplet(const int base, const int top, const int stride = 1) const {
    int extent ;
    if((stride > 0 && top < base) || (stride < 0 && top > base))
      extent = 0 ;
    else {
      const int parExtent = size() ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: trip: subrange out of range") ;
      else
        extent = (top - base) / stride + 1 ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  // Inquiries

  int size() const {return rep->size() ; }

  Dimension dim() const {return rep->dim() ; }

  Range ker() const { return Range(rep->ker()) ; }
  Range shell() const { return Range(rep->shell()) ; }

  Range tem() const { return Range(rep->tem()) ; }
  int bas() const {return rep->bas() ; }
  int str() const {return rep->str() ; }

  Format format() const {return rep->format() ; }

  // Subscripting and iteration

  void location(Location* loc, const int glb) const {
    rep->location(loc, glb) ;
  }

  void block(Block* blk, const int ker_glb) const {
    rep->block(blk, ker_glb) ;
  }

  void block(Block* blk, const int ker_glb,
             const int count, const int base) const {
    rep->block2(blk, ker_glb, count, base) ;
  }

  void block(Block* blk, const int ker_glb,
             const int count, const int base, const int stride) const {
    if(stride == 1)
      rep->block2(blk, ker_glb, count, base) ;
    else
      rep->block3(blk, ker_glb, count, base, stride) ;
  }

//  int local(int* glb) const {return rep->local(glb) ; }

  void crds(int* ker_count, int* ker_bas, int* ker_str) const {
    rep->crds(ker_count, ker_bas, ker_str) ;
  }

  void crds(int* ker_count, int* ker_bas, int* ker_str,
               const int count, const int base) const {
    rep->crds2(ker_count, ker_bas, ker_str, count, base) ;
  }

  void crds(int* ker_count, int* ker_bas, int* ker_str,
               const int count, const int base, const int stride) const {
    if(stride == 1)
    rep->crds2(ker_count, ker_bas, ker_str, count, base) ;
    else
      rep->crds3(ker_count, ker_bas, ker_str, count, base, stride) ;
  }

  // `Map' stuff

  int disp(const int sub) const {return rep->disp(sub) ; }
  int step(const int sub_stp) const {return rep->step(sub_stp) ; }

  int offset(const Location& i) const {return rep->offset(&i) ; }

  int volume() const {return rep->volume() ; }

  // Alignment and range comparisions

  int isAligned(const Range& x) {
    return rep->isAligned(x.rep) ;
  }

  // Miscellaneous operators

  Location operator()(const int glb) {
    Location res;
    location(&res, glb) ;
    return res ;
  }

  Range operator()(const Trip& t) {
    const int base   = t.lb ;
    const int top    = t.ub != DEF_UB ? t.ub : size() - 1 ;
    const int stride = t.str ;
    return triplet(base, top, stride) ;
  }

  Range& operator=(const Range& x) {
    if(x.rep) x.rep->incr() ;
    if(rep)   rep->decr() ;
    rep = x.rep ;
    return *this ;
  }

  // Interim support for PCRC Fortran interface

  int glb(const int sub, const int ker_glb) const {
    return rep->glb(sub, ker_glb) ;
  }

  int& g_lb() {return rep->g_lb ; }
  int& g_stride() {return rep->g_stride ; }

  // debugging

  void show() {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Range :: show: Null range\n") ;
  }
protected :
  Range(RangeRep* _rep) : rep(_rep) {
    if(rep) rep->incr() ;
  }

  RangeRep* rep ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRange': collapsed range.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollapsedRange : public Range {
public :
  CollapsedRange(const int extent) {
    rep = new CollapsedRangeRep(extent) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `DimRange': range representing a process dimension.           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimRange : public Range {
public :
  DimRange(Dimension dim) {
    rep = new DimRangeRep(dim) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRange': block-distributed range.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRange : public Range {
public :
  BlockRange(const int extent, Dimension dim) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new BlockRangeRep(extent, new DimRangeRep(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int wlo, const int whi) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new ExtendedBlockRangeRep(extent, new DimRangeRep(dim), blockSize,
                                    wlo, whi) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim,  const int blockSize) {
    rep = new BlockRangeRep(extent, new DimRangeRep(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int blockSize,
	     const int wlo, const int whi) {
    rep = new ExtendedBlockRangeRep(extent, new DimRangeRep(dim), blockSize,
                                    wlo, whi) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRange': cyclic-distributed range.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRange : public Range {
public :
  CyclicRange(const int extent, Dimension dim) {
    const int numCycles = (extent + dim.size() - 1) / dim.size() ;
    rep = new CyclicRangeRep(extent, new DimRangeRep(dim), numCycles) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRange': block-cyclically distributed range.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRange : public Range {
public :
  BlockCyclicRange(const int extent, Dimension dim, const int blockSize) {
    const int numBlocks = (extent + blockSize - 1) / blockSize ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;
    RangeRep* ker = new CyclicRangeRep(numBlocks, new DimRangeRep(dim),
                                       numCycles) ;
    rep = new BlockCyclicRangeRep(extent, ker, blockSize) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregRange': irregular block-distributed range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class IrregRange : public Range {
public :
  IrregRange(Dimension dim, int* blocks) {
    rep = new IrregRangeRep(new DimRangeRep(dim), blocks) ;
    rep->incr() ;
  }

  IrregRange(const int extent, Dimension dim, Integer* map) {
    rep = new IrregRangeRep(extent, new DimRangeRep(dim), map) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRange': stepped block-distributed range.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRange : public Range {
public :
  StepRange(const int extent, Dimension dim) {
    rep = new StepRangeRep(new DimRangeRep(dim),
                           extent / dim.size(), extent % dim.size()) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockRange': block-distributed range yielding arrays    //
//                          with special packed memory layout          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockRange : public Range {
public :
  ADAPTBlockRange(const int extent, const int base, const int stride,
                  const int parExtent, Dimension dim, const int blockSize) {

    if(extent != parExtent || stride < 0) {
      BlockRangeRep* par =
          new PackedBlockRangeRep(parExtent, new DimRangeRep(dim),
			          blockSize, stride > 0 ? stride : -stride) ;

      rep = par->subrng(extent, base, stride) ;
    }
    else
      rep = new BlockRangeRep(parExtent, new DimRangeRep(dim), blockSize) ;

    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTCyclicRange': cyclic-distributed range yielding arrays  //
//                           with special truncated memory layout.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTCyclicRange : public Range {
public :
  ADAPTCyclicRange(const int extent, const int base, const int stride,
                   Dimension dim) {
    const int parExtent =
        (stride > 0 ? base + stride * (extent - 1) : base) + 1 ;
    const int numCycles = (parExtent + dim.size() - 1) / dim.size() ;

    RangeRep* par = new CyclicRangeRep(parExtent, new DimRangeRep(dim),
                                       numCycles) ;

    rep = (extent != par->size() || stride < 0) ?
              par->subrng(extent, base, stride) : par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockCyclicRange': block-cyclically distributed range,  //
//                                yielding arrays with special packed  //
//                                memory layout.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockCyclicRange : public Range {
public :
  ADAPTBlockCyclicRange(const int extent, const int base, const int stride,
                        Dimension dim, const int blockSize) {
    const int numBlocks =
        (stride > 0 ? base + stride * (extent - 1) : base) / blockSize + 1 ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;

    RangeRep* ker = new CyclicRangeRep(numBlocks, new DimRangeRep(dim),
                                       numCycles) ;

    BlockRangeRep* par = new BlockCyclicRangeRep(numBlocks * blockSize,
                                                 ker, blockSize) ;

    if(extent != par->size() || stride < 0) {
      BlockRangeRep* par =
          new PackedBlockCyclicRangeRep(numBlocks * blockSize, ker, blockSize,
                                        stride > 0 ? stride : -stride) ;

      rep = par->subrng(extent, base, stride) ;
    }
    else
      rep = new BlockCyclicRangeRep(numBlocks * blockSize, ker, blockSize) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Map'                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Map {
public :
  Map() : stride(0) {}

  Map(Range _range, const int _stride) :
               range(_range), stride(_stride) {}

  Map ker() const {
    return Map(range.ker(), range.shell().volume() * stride) ;
  }

  Map shell() const {
    return Map(range.shell(), stride) ;
  }

  Range rng() const {return range ; }

  int str() const {return stride ; }

  int offset(const Location& i) const {
    stride * range.offset(i) ;
    return stride * range.offset(i) ;
  }

  int disp(const int sub) const {
    return stride * range.disp(sub) ;
  }

  int step(const int sub_stp) const {
    return stride * range.step(sub_stp) ;
  }

  Range range ;
  int stride ;
} ;


// Iterator (index) classes

inline int inRange(const int val,
                   const int count, const int base, const int stride) {
  const int offset = val - base ;
  const int pos    = offset / stride ;
  return pos >= 0 && pos < count && pos * stride == offset ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex': enumerate local blocks of range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex : public Location, public Block {
public :
  LocBlocksIndex(Range _x) : x(_x), simple(_x.format() != DIST_BLOCK_CYCLIC) {

    x.crds(&d_count, &d_bas, &d_str) ;

    dim = x.dim() ;
  }

  void beginLocBlk() {

    crd = dim.crd() ;

    looping = 0 ;

    if(simple) {  // Normal case

      if(inRange(crd, d_count, d_bas, d_str)) {
        x.block(this, crd) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

    else {  // Block-Cyclic case

      // Here the triplet range (d_count, d_bas, d_str) is in terms of
      // ``virtual processors'', ie, relative to the kernel cyclic
      // range rather than the actual process dimension.
      //
      // `d_str' may be +1 or -1.

      P = dim.size() ;

      // Compute local range of `blk' subscript, which
      // identifies individual blocks.  Kernel subscript
      // is invariantly:
      //
      //           ker  =  crd + P * blk
      //
      // (where `ker' is in the `d'-triplet range).

      if(d_str > 0) {
        blk       = (d_bas - crd + P - 1) / P ;
        blk_count = (d_bas + d_count - crd + P - 1) / P - blk ;

        ker       = crd + P * blk ;
        ker_stp   = P ;
      }
      else {
        blk       = (d_bas - crd + P) / P - 1 ;
        blk_count = blk - (d_bas - d_count - crd + P) / P + 1 ;

        ker       = crd + P * blk ;
        ker_stp   = -P ;
      }

      blk_l = 0 ;
      if(blk_l < blk_count) {
        x.block(this, ker) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }
  }

  int test() {return looping ; }

  void nextLocBlk() {

    if(simple)  // Normal case
      looping = 0 ;

    else {  // Block-Cyclic case

      blk += d_str ;
      ker += ker_stp ;

      blk_l++ ;
      if(blk_l < blk_count) {
        x.block(this, ker) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

protected :
  Range x ;
  int simple ;  // Flag set TRUE if distribution is *not* block-cyclic.

  int d_count, d_bas, d_str ;

  int looping ;

  // For block cyclic ranges:

  int P ;

  int blk_count, blk_l, ker, ker_stp ;

private :
  LocBlocksIndex(const LocBlocksIndex&) {} ;  // No copying allowed.
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex2': Like `LocBlocksIndex', but range           //
//                          restricted by an interval.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex2 : public Location, public Block {
public :
  LocBlocksIndex2(Range _x,
                  const int _i_count, const int _i_bas) :
        x(_x), i_count(_i_count), i_bas(_i_bas),
        simple(_x.format() != DIST_BLOCK_CYCLIC) {

    x.crds(&d_count, &d_bas, &d_str, i_count, i_bas) ;

    dim = x.dim() ;
  }

  void beginLocBlk() {

    crd = dim.crd() ;

    looping = 0 ;

    if(simple) {  // Normal case

      if(inRange(crd, d_count, d_bas, d_str)) {

        x.block(this, crd, i_count, i_bas) ;

        sub = sub_bas ;

        looping = 1 ;
      }

    }

    else {  // Block-Cyclic case

      // Here the triplet range (d_count, d_bas, d_str) is in terms of
      // ``virtual processors'', ie, relative to the kernel cyclic
      // range rather than the actual process dimension.
      //
      // `d_str' may be +1 or -1.

      P = dim.size() ;

      // Compute local range of `blk' subscript, which
      // identifies individual blocks.  Kernel subscript
      // is invariantly:
      //
      //           ker  =  crd + P * blk
      //
      // (where `ker' is in the `d'-triplet range).

      if(d_str > 0) {
        blk       = (d_bas - crd + P - 1) / P ;
        blk_count = (d_bas + d_count - crd + P - 1) / P - blk ;

        ker       = crd + P * blk ;
        ker_stp   = P ;
      }
      else {
        blk       = (d_bas - crd + P) / P - 1 ;
        blk_count = blk - (d_bas - d_count - crd + P) / P + 1 ;

        ker       = crd + P * blk ;
        ker_stp   = -P ;
      }

      blk_l = 0 ;
      if(blk_l < blk_count) {
        x.block(this, ker, i_count, i_bas) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }
  }

  int test() {return looping ; }

  void nextLocBlk() {

    if(simple)  // Normal case
      looping = 0 ;

    else {  // Block-Cyclic case

      blk += d_str ;
      ker += ker_stp ;

      blk_l++ ;
      if(blk_l < blk_count) {
        x.block(this, ker, i_count, i_bas) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

public:
  int i_count, i_bas ;

protected :
  Range x ;
  int simple ;  // Flag set TRUE if distribution is *not* block-cyclic.

  int d_count, d_bas, d_str ;

  int looping ;

  // For block cyclic ranges:

  int P ;

  int blk_count, blk_l, ker, ker_stp ;

private :
  LocBlocksIndex2(const LocBlocksIndex2&) {} ;  // No copying allowed.
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex3': Like `LocBlocksIndex', but range           //
//                          restricted by a triplet.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex3 : public Location, public Block {
public :
  LocBlocksIndex3(Range _x,
                  const int _i_count, const int _i_bas, const int _i_str) :
        x(_x), i_count(_i_count), i_bas(_i_bas), i_str(_i_str),
        simple(_x.format() != DIST_BLOCK_CYCLIC) {

    x.crds(&d_count, &d_bas, &d_str, i_count, i_bas, i_str) ;

    dim = x.dim() ;
  }

  void beginLocBlk() {

    crd = dim.crd() ;

    looping = 0 ;

    if(simple) {  // Normal case

      if(inRange(crd, d_count, d_bas, d_str)) {

        x.block(this, crd, i_count, i_bas, i_str) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

    else {  // Block-Cyclic case

      // Here the triplet range (d_count, d_bas, d_str) is in terms of
      // ``virtual processors'', ie, relative to the kernel cyclic
      // range rather than the actual process dimension.
      //
      // `d_str' may be +1 or -1.

      P = dim.size() ;

      // Compute local range of `blk' subscript, which
      // identifies individual blocks.  Kernel subscript
      // is invariantly:
      //
      //           ker  =  crd + P * blk
      //
      // (where `ker' is in the `d'-triplet range).

      if(d_str > 0) {
        blk       = (d_bas - crd + P - 1) / P ;
        blk_count = (d_bas + d_count - crd + P - 1) / P - blk ;

        ker       = crd + P * blk ;
        ker_stp   = P ;
      }
      else {
        blk       = (d_bas - crd + P) / P - 1 ;
        blk_count = blk - (d_bas - d_count - crd + P) / P + 1 ;

        ker       = crd + P * blk ;
        ker_stp   = -P ;
      }

      blk_l = 0 ;
      if(blk_l < blk_count) {
        x.block(this, ker, i_count, i_bas, i_str) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }
  }

  int test() {return looping ; }

  void nextLocBlk() {

    if(simple)  // Normal case
      looping = 0 ;

    else {  // Block-Cyclic case

      blk += d_str ;
      ker += ker_stp ;

      blk_l++ ;
      if(blk_l < blk_count) {
        x.block(this, ker, i_count, i_bas, i_str) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

public :
  int i_count, i_bas, i_str ;

protected :
  Range x ;
  int simple ;  // Flag set TRUE if distribution is *not* block-cyclic.

  int d_count, d_bas, d_str ;

  int looping ;

  // For block cyclic ranges:

  int P ;

  int blk_count, blk_l, ker, ker_stp ;

private :
  LocBlocksIndex3(const LocBlocksIndex3&) {} ;  // No copying allowed.
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex': enumerate all blocks of range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex : public LocBlocksIndex {
public :
  AllBlocksIndex(Range _x) : LocBlocksIndex(_x) {}

  void beginAllBlk() {

    looping = 0 ;

    if(simple) {  // Normal case

      crd = d_bas ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, crd) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

    else {  // Block-Cylic case

      // The triplet (d_count, d_bas, d_str) defines range of
      // subscripts in kernel cyclic range.

      P = dim.size() ;

      d_l = 0 ;

      ker = d_bas ;

      crd = ker % P ;
      blk = ker / P ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, ker) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

  }

  void nextAllBlk() {
    if(simple) {  // Normal case

      d_l++ ;
      crd += d_str ;

      if(d_l < d_count) {
        x.block(this, crd) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
 
    else {  // Block-Cyclic case

      d_l++ ;
      ker += d_str ;

      crd = ker % P ;
      blk = ker / P ;

      if(d_l < d_count) {
        x.block(this, ker) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

private :

  int d_l ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex2': Like `AllBlocksIndex', but range           //
//                          restricted by an interval.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex2 : public LocBlocksIndex2 {
public :
  AllBlocksIndex2(Range _x,
                  const int _i_count, const int _i_bas) :
                      LocBlocksIndex2(_x, _i_count, _i_bas) {}

  void beginAllBlk() {

    looping = 0 ;

    if(simple) {  // Normal case

      crd = d_bas ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, crd, i_count, i_bas) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

    else {  // Block-Cylic case

      // The triplet (d_count, d_bas, d_str) defines range of
      // subscripts in kernel cyclic range.

      P = dim.size() ;

      d_l = 0 ;

      ker = d_bas ;

      crd = ker % P ;
      blk = ker / P ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, ker, i_count, i_bas) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

  }

  void nextAllBlk() {
    if(simple) {  // Normal case

      d_l++ ;
      crd += d_str ;

      if(d_l < d_count) {
        x.block(this, crd, i_count, i_bas) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
 
    else {  // Block-Cyclic case

      d_l++ ;
      ker += d_str ;

      crd = ker % P ;
      blk = ker / P ;

      if(d_l < d_count) {
        x.block(this, ker, i_count, i_bas) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

private :

  int d_l ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex3': Like `AllBlocksIndex', but range           //
//                          restricted by a triplet.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex3 : public LocBlocksIndex3 {
public :
  AllBlocksIndex3(Range _x,
                  const int _i_count, const int _i_bas, const int _i_str) :
                      LocBlocksIndex3(_x, _i_count, _i_bas, _i_str) {}

  void beginAllBlk() {

    looping = 0 ;

    if(simple) {  // Normal case

      crd = d_bas ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, crd, i_count, i_bas, i_str) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

    else {  // Block-Cylic case

      // The triplet (d_count, d_bas, d_str) defines range of
      // subscripts in kernel cyclic range.

      P = dim.size() ;

      d_l = 0 ;

      ker = d_bas ;

      crd = ker % P ;
      blk = ker / P ;

      d_l = 0 ;
      if(d_l < d_count) {
        x.block(this, ker, i_count, i_bas, i_str) ;

        sub = sub_bas ;

        looping = 1 ;
      }
    }

  }

  void nextAllBlk() {
    if(simple) {  // Normal case

      d_l++ ;
      crd += d_str ;

      if(d_l < d_count) {
        x.block(this, crd, i_count, i_bas, i_str) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
 
    else {  // Block-Cyclic case

      d_l++ ;
      ker += d_str ;

      crd = ker % P ;
      blk = ker / P ;

      if(d_l < d_count) {
        x.block(this, ker, i_count, i_bas, i_str) ;

        sub = sub_bas ;
      }
      else
        looping = 0 ;
    }
  }

private :

  int d_l ;
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index'.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index : public LocBlocksIndex {
public :
  Index(Range _x) : LocBlocksIndex(_x) {}

  void begin() {
    beginLocBlk() ;
    while(count == 0 && looping)
      nextLocBlk() ;

    idx = glb_bas ;

    l   = 0 ;
  }

  void next() {
    sub += sub_stp ;

    idx += glb_stp ;

    l++ ;
    if(l == count) {
      nextLocBlk() ;
      while(count == 0 && looping)
	nextLocBlk() ;

      idx = glb_bas ;

      l   = 0 ;
    }
  }

  operator int() const { return idx ; }

private :
  int l ;
  int idx ;

} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index2'.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index2 : public LocBlocksIndex2 {
public :
  Index2(Range _x, const int _i_count, const int _i_bas) :
                     LocBlocksIndex2(_x, _i_count, _i_bas) {}

  void begin() {
    beginLocBlk() ;
    while(count == 0 && looping)
      nextLocBlk() ;

    idx = i_bas + glb_bas ;

    l   = 0 ;
  }

  void next() {
    sub += sub_stp ;

    idx += glb_stp ;

    l++ ;
    if(l == count) {
      nextLocBlk() ;
      while(count == 0 && looping)
	nextLocBlk() ;

      idx = i_bas + glb_bas ;

      l   = 0 ;
    }
  }

  operator int() const { return idx ; }

private :
  int l ;
  int idx ;

} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index3'.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index3 : public LocBlocksIndex3 {
public :
  Index3(Range _x, const int _i_count, const int _i_bas, const int _i_str) :
                     LocBlocksIndex3(_x, _i_count, _i_bas, _i_str) {}

  void begin() {
    beginLocBlk() ;
    while(count == 0 && looping)
      nextLocBlk() ;

    idx = i_bas + i_str * glb_bas ;

    l   = 0 ;
  }

  void next() {
    sub += sub_stp ;

    idx += i_str * glb_stp ;

    l++ ;
    if(l == count) {
      nextLocBlk() ;
      while(count == 0 && looping)
	nextLocBlk() ;

      idx = i_bas + i_str * glb_bas ;

      l   = 0 ;
    }
  }

  operator int() const { return idx ; }

private :
  int l ;
  int idx ;

} ;


// Cosmetic support for ghost regions:

class IndexedLocation : public Location {
public :
  IndexedLocation(const Location& i, const int _idx) :
                  Location(i), idx(_idx) {}

  operator int() { return idx ; }

private :
  int idx ;
} ;

inline IndexedLocation operator+(const Index& i, const int shift) {
  IndexedLocation j(i, (int) i + shift) ;
  j.sub += shift * i.sub_stp ;
  return j ;
}

inline IndexedLocation operator-(const Index& i, const int shift) {
  IndexedLocation j(i, (int) i - shift) ;
  j.sub -= shift * i.sub_stp ;
  return j ;
}

inline IndexedLocation operator+(const Index2& i, const int shift) {
  IndexedLocation j(i, (int) i + shift) ;
  j.sub += shift * i.sub_stp ;
  return j ;
}

inline IndexedLocation operator-(const Index2& i, const int shift) {
  IndexedLocation j(i, (int) i - shift) ;
  j.sub -= shift * i.sub_stp ;
  return j ;
}

inline IndexedLocation operator+(const Index3& i, const int shift) {
  IndexedLocation j(i, (int) i + shift) ;
  j.sub += shift * i.sub_stp / i.i_str ;   // Unfortunate case.
  return j ;
}

inline IndexedLocation operator-(const Index3& i, const int shift) {
  IndexedLocation j(i, (int) i - shift) ;
  j.sub -= shift * i.sub_stp / i.i_str ;   // Unfortunate case.
  return j ;
}



// ---------------------------------------------------------------
//
// Things to do:
//
//   Change name of `Range :: volume' to `Range :: domain'?
//
//   Change name of `Range :: disp' to `Range :: offset' (ie,
//       overlead the name `offset')?
//
//   Add a `shf_stp' field to `Index', `Index2', `Index3', and
//       initialize it by a new inquiry on `Range'.  Used in the
//       shift operations to avoid the ``unfortunate case''.

