
// Schedules for irregular access to remote data.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "threads.h"
#include "adtags.h"
#include "ProcTree.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "DataSchedule.h"

#ifdef USE_ARMCI
#include "armci.h"
#endif

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `AckServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Auxilliary class used by (MPI versions of) `DataSchedule' and
// `BlockDataSchedule'.

class AckServer : public Server {
public :
  AckServer () ;
  ~AckServer () {}

  void transition(MPI_Status* status) ;

  void notify() ;
  int waiting() ;
private :
  int m ;
  int dummy ;
} ;


AckServer :: AckServer() {
  m = 0 ;
}

void AckServer :: transition(MPI_Status*) {
  m-- ;
  if(m > 0) {
    // enable receipt of next acknowledgement
    MPI_Irecv(&dummy, 1, MPI_INT,
	      MPI_ANY_SOURCE, ACK, MPI_COMM_WORLD, event) ;
  }
}

void AckServer :: notify() {
  // Another `ack' is expected.

  if(m == 0) {
    // enable receipt of acknowledgement
    MPI_Irecv(&dummy, 1, MPI_INT,
	      MPI_ANY_SOURCE, ACK, MPI_COMM_WORLD, event) ;
  }

  m++ ;
}

int AckServer :: waiting() {
  return m ;
}



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `DataSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#ifdef USE_ARMCI  // Implementation based on ARMCI.


DataSchedule :: DataSchedule(const int _elementLen) : elementLen(_elementLen) {
  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

  locIdxs = new IndexList [nProcs] ;
  remIdxs = new IndexList [nProcs] ;
}

void DataSchedule :: build(TreeSchedule* spanTree) {

  ...
}

void gather(void* locBase, void* srvBase) {

  spanTree->synch() ;

  for(int remId = 0 ; remId < nProcs ; remId++) {

    IndexList* remVec = remIdxs + remId ;
    IndexList* locVec = locIdxs + remId ;

    const int n = remVec->size() ;
    if(n > 0) {

      armci_giov_t desc ;
      desc.src_ptr_ar = remVec->data() ;
      desc.dst_ptr_ar = locVec->data() ;
      desc.bytes      = elementLen ;
      desc.ptr_ar_len = n ;

      ARMCI_GetV(&desc, 1, remId) ;
    }
  }

  spanTree->synch() ;
}


DataSchedule :: ~DataSchedule() {
  delete [] remIdxs ;
  delete [] locIdxs ;
}


#else  // Direct MPI implementation.


// `RemServer' class services remote access requests, building up the
// `srvIdxs' list.

class RemServer : public Server {
public :
  RemServer(IndexList* _srvIdxs) ;
  ~RemServer() ;

  void transition(MPI_Status*) ;
private :
  IndexList* srvIdxs ;

  enum {ALPHA_0, ALPHA_1, ALPHA_2} ;
  int state ;
  int n, srcId ;

  int dummy ;
} ;


DataSchedule :: DataSchedule(const int _elementLen) :
                    MessSchedule(_elementLen) {
  remIdxs = new IndexList [nProcs] ;
}

void DataSchedule :: build(TreeSchedule* spanTree) {

  // Communication to build the `srvIdxs' vectors from the `remIdxs' vectors.

  RemServer remServer(srvIdxs) ;
  AckServer ackServer ;
  spanTree->synch() ;

  for(int hop = 0 ; hop < nProcs ; hop++) {
    const int remId = (locId + hop) % nProcs ;

    IndexList* idxVec = remIdxs + remId ;

    int n = idxVec->size() ;
    if(n > 0) {
      int* offsets = idxVec->data() ;

      // Send offsets to `remServer' on remote process.

      MPI_Status status ;

      MPI_Isend(&n, 1, MPI_INT, remId, REM, MPI_COMM_WORLD, event) ; 
      Server :: waitClientEvent(&status) ;

      MPI_Isend(offsets, n, MPI_INT, remId, DAT, MPI_COMM_WORLD, event) ; 
      Server :: waitClientEvent(&status) ;

      ackServer.notify() ;
    }
  }

  while(ackServer.waiting())
    Server :: waitServerTransition() ;
  spanTree->synch() ;

  delete [] remIdxs ;  // potentially move this into destructor.

  MessSchedule :: build() ;
}


RemServer :: RemServer(IndexList* _srvIdxs) : srvIdxs(_srvIdxs) {

  // enable receipt of first remote access request
  MPI_Irecv(&n, 1, MPI_INT, MPI_ANY_SOURCE, REM, MPI_COMM_WORLD, event) ;

  state = ALPHA_0 ;
}

void RemServer :: transition(MPI_Status* status) {
  switch(state) {
    case ALPHA_0 : {
      srcId = status->MPI_SOURCE ;

      IndexList* idxVec = srvIdxs + srcId ;

      // Allocate and enable receipt of ``offsets'' vector.

      int* offsets = idxVec->addSome(n) ;
      MPI_Irecv(offsets, n, MPI_INT, srcId, DAT, MPI_COMM_WORLD, event) ;

      state = ALPHA_1 ;
      break ;
    }
    case ALPHA_1 : {
      // enable send of acknowledgement
      MPI_Isend(&dummy, 1, MPI_INT, srcId, ACK, MPI_COMM_WORLD, event) ;

      state = ALPHA_2 ;
      break ;
    }
    case ALPHA_2 : {
      // enable receipt of next remote access request
      MPI_Irecv(&n, 1, MPI_INT, MPI_ANY_SOURCE, REM, MPI_COMM_WORLD, event) ;

      state = ALPHA_0 ;
      break ;
    }
  }
}

RemServer :: ~RemServer() {
  // Terminate pending communications

  if(state == ALPHA_0) {
    // Server is waiting for a request.  Send one to self.

    // (It is also allowed for the server to be in state
    // ALPHA_2---waiting for a send to complete---at the time of
    // termination.)

    int locId ;
    int dummyN ;
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Send(&dummyN, 1, MPI_INT, locId, REM, MPI_COMM_WORLD) ;
  }

  MPI_Status status ;
  MPI_Wait(event, &status) ;
}

#endif

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockDataSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// `BlockRemServer' class services remote access requests, building up the
// `srvMessages' list.

class BlockRemServer : public Server, public GlobalEnv {
public :
  BlockRemServer(SparseArray<MessQueue>* _srvMessages) ;
  ~BlockRemServer() ;

  void transition(MPI_Status*) ;
private :
  SparseArray<MessQueue>* srvMessages ;

  enum {ALPHA_0, ALPHA_1, ALPHA_2, ALPHA_3} ;
  int state ;

  int header [2] ;

  int srcId ;

  int n ;
  int* blocks ;

  int dummy ;
} ;


void BlockDataSchedule :: sendRequests(TreeSchedule* spanTree) {

  // Communication to build the `srvMessages' array from the
  // `remMessages' array.

  BlockRemServer remServer(&srvMessages) ;
  AckServer ackServer ;
  spanTree->synch() ;

  List<MessQueue>* record = remMessages->toList() ;

  while(record) {
    const int remId    = record->remId ;

    const int numMess  = record->messages.size() ;
    const int n        = record->blocks.size() ;

    int header [2] ;
    header [0] = numMess ;
    header [1] = n ;

    // Send blocks to `remServer' on remote process.

    MPI_Status status ;

    MPI_Isend(header, 2, MPI_INT, remId, REM, MPI_COMM_WORLD, event) ; 
    Server :: waitClientEvent(&status) ;

    MPI_Isend(record->messages.data(), numMess * sizeof(Mess), MPI_CHAR, remId,
	      DAT, MPI_COMM_WORLD, event) ; 
    Server :: waitClientEvent(&status) ;

    MPI_Isend(record->blocks.data(), n, MPI_INT, remId,
	      DAT, MPI_COMM_WORLD, event) ; 
    Server :: waitClientEvent(&status) ;

    ackServer.notify() ;

    record = record->next ;
  }

  while(ackServer.waiting())
    Server :: waitServerTransition() ;

  spanTree->synch() ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRemServer' class.                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Services remote access requests, building up the `srvMessages' array.

BlockRemServer :: BlockRemServer(SparseArray<MessQueue>* _srvMessages) :
                      srvMessages(_srvMessages) {

  // enable receipt of first remote access request
  MPI_Irecv(header, 2, MPI_INT, MPI_ANY_SOURCE, REM, MPI_COMM_WORLD, event) ;

  state = ALPHA_0 ;
}

BlockRemServer :: ~BlockRemServer() {
  // Terminate pending communications

  if(state == ALPHA_0) {
    // Server is waiting for a request.  Send one to self.

    // (It is also allowed for the server to be in state
    // ALPHA_3---waiting for a send to complete---at the time of
    // termination.)

    int dummyHeader [2] ;
    MPI_Send(dummyHeader, 2, MPI_INT, locId, REM, MPI_COMM_WORLD) ;
  }

  MPI_Status status ;
  MPI_Wait(event, &status) ;
}


void BlockRemServer :: transition(MPI_Status* status) {
  switch(state) {
    case ALPHA_0 : {
      srcId = status->MPI_SOURCE ;

      const int numMess = header [0] ;
      n                 = header [1] ;

      const int hop = (srcId - locId + nProcs) % nProcs ;
      MessQueue* queue = srvMessages->getElement(hop) ;
      queue->remId = srcId ;
      queue->hop   = hop   ;

      Mess* messages = queue->messages.addSome(numMess) ;
      blocks         = queue->blocks.addSome(n) ;

      // Enable receipt of `messages' vector.

      MPI_Irecv(messages, numMess * sizeof(Mess), MPI_CHAR,
                srcId, DAT, MPI_COMM_WORLD, event) ;

      state = ALPHA_1 ;
      break ;
    }
    case ALPHA_1 : {
      // Enable receipt of `blocks' vector.

      MPI_Irecv(blocks, n, MPI_INT, srcId, DAT, MPI_COMM_WORLD, event) ;

      state = ALPHA_2 ;
      break ;
    }
    case ALPHA_2 : {
      // Enable send of acknowledgement

      MPI_Isend(&dummy, 1, MPI_INT, srcId, ACK, MPI_COMM_WORLD, event) ;

      state = ALPHA_3 ;
      break ;
    }
    case ALPHA_3 : {
      // Enable receipt of next remote access request

      MPI_Irecv(header, 2, MPI_INT, MPI_ANY_SOURCE,
                REM, MPI_COMM_WORLD, event) ;

      state = ALPHA_0 ;
      break ;
    }
  }
}


// ----------------------------------------------------------------------
// Things to do
//
//   Get rid of calls to blocking send, `MPI_Send'?  No real reason for
//       this---just that these operations aren't used anywhere except in
//       the destructors of the `remServer's to send a message to self.
//       It would somehow be nice to say I never use blocking communications...
//
//   Find out why are the `RemServer' destructors are using direct calls to
//       `MPI_wait'?!

