
// Schedules for irregular subscripting of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "DataSchedule.h"
#include "Reduce.h"
#include "Gather.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherSkeleton'                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

GatherSkeleton :: GatherSkeleton(const DAD* _loc, const DAD* _rem,
                                 const int len,
                                 const DAD** _subs, const int** subsData,
                                 const DAD* _msk, const int* mskData) :
      DataSchedule(len),
      loc(_loc), rem(_rem), subs(_subs), msk(_msk), rank(_loc->rnk()) {

  Group p = loc->grp() ;
  q = rem->grp() ;

  for(int d = 0 ; d < rank ; d++)
    alignCheck(loc, subs [d]) ;
  if(msk)
    alignCheck(loc, msk) ;

  if(!apg.contains(p))
    AdlibError("GatherSkeleton :: GatherSkeleton: loc is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("GatherSkeleton :: GatherSkeleton: rem is inaccessible.\n") ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, subsData, mskData, 0) ;
  }

  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}

void GatherSkeleton :: loop(int locOffset,
                            const int** subsElem, const int* mskElem,
                            const int r) {

  if(r == rank) {
    if(!msk || *mskElem) {
      Group  remGrp = q ;
      int remOffset = 0 ;

      for(int s = 0 ; s < rem->rnk() ; s++) {
	Range y(rem->rng(s)) ;
	Location i = y(*(subsElem [s])) ;

        remOffset += rem->map(s).offset(i) ;
        //remGrp    /= i ;
        remGrp     = remGrp / i ;
      }

      transReq(locOffset, remGrp.lead_abs(), remOffset) ;
    }
  }
  else {
    LocBlocksIndex i(loc->rng(r)) ;

    Map locMap = loc->map(r) ;
    Map mskMap ;
    if(msk) mskMap = msk->map(r) ;

    const int** subsDataElem = new const int* [rem->rnk()] ;
    int* subsStep = new int [rem->rnk()] ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int s ;

      int locOffsetElem = locOffset + locMap.offset(i) ;
      const int locStep = locMap.step(i.sub_stp) ;

      for(s = 0 ; s < rem->rnk() ; s++) {
        Map subMap = (subs [s])->map(r) ;

        subsDataElem [s] = subsElem [s] + subMap.offset(i) ;
        subsStep [s]     = subMap.step(i.sub_stp) ;
      }

      const int* mskDataElem = msk ? mskElem + mskMap.offset(i) : (int*) NULL ;
      const int mskStep = msk ? mskMap.step(i.sub_stp) : 0 ;

      for(int i_block = 0 ; i_block < i.count ; i_block++) {
        loop(locOffsetElem, subsDataElem, mskDataElem, r + 1) ;

	locOffsetElem += locStep ;
        for(s = 0 ; s < rem->rnk() ; s++)
          subsDataElem [s] += subsStep [s] ;
        if(msk)
          mskDataElem += mskStep ;
      }
    }

    delete [] subsStep ;
    delete [] subsDataElem ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapeSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapeSkeleton :: ReshapeSkeleton(const DAD* _loc, const DAD* _rem,
                                   const int len) :
		       loc(_loc), rem(_rem), rank(_loc->rnk()),
		       DataSchedule(len) {

  Group p = loc->grp() ;
  q = rem->grp() ;

  if(!apg.contains(p))
    AdlibError("ReshapeSkeleton :: ReshapeSkeleton: loc is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("ReshapeSkeleton :: ReshapeSkeleton: rem is inaccessible.\n") ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, 0, 1, 0) ;
  }

  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}


void ReshapeSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                             int r) {

  if(r == rank) {
    Group remGrp = q ;
    int remOffset = 0 ;

    for(int s = 0 ; s < rem->rnk() ; s++) {
      Range y(rem->rng(s)) ;
      Location i = y(seqIndex % y.size()) ;

      remOffset += rem->map(s).offset(i) ;
      //remGrp    /= i ;
      remGrp     = remGrp / i ;
    }

    transReq(locOffset, remGrp.lead_abs(), remOffset) ;
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.begin() ; i.test() ; i.next()) {
      int locOffsetElem = locOffset + loc->map(r).offset(i) ;

      loop(locOffsetElem,
//           seqIndex + seqStride * x.idx(i), x.size() * seqStride,
           seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePadSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapePadSkeleton :: ReshapePadSkeleton(const DAD* _loc, int _remSize,
                                         const DAD* _pad,
                                         const int len) :
			  loc(_loc), remSize(_remSize), pad(_pad),
                          rank(_loc->rnk()),
			  DataSchedule(len) {

  // Silly `reshape' padding to support Fortran.

  Group p = loc->grp() ;
  q = pad->grp() ;

  if(!apg.contains(p))
    AdlibError("ReshapePadSkeleton :: ReshapePadSkeleton: "
               "loc is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("ReshapePadSkeleton :: ReshapePadSkeleton: "
               "pad is inaccessible.\n") ;


  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, 0, 1, 0) ;
  }

  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}

void ReshapePadSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                                int r) {

  if(r == rank) {
    if(seqIndex >= remSize) {
      Range y(pad->rng(0)) ;
      const int padIndex = (seqIndex - remSize) % y.size() ;
 
      Location i = y(padIndex) ;

      Group remGrp = q / i ;
      transReq(locOffset, remGrp.lead_abs(), pad->map(0).offset(i)) ;
    }
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.begin() ; i.test() ; i.next()) {
      loop(locOffset + loc->map(r).offset(i),
//	   seqIndex + seqStride * x.idx(i), x.size() * seqStride,
	   seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherSkeleton'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


int VecGatherSkeleton :: blockRank(const int rank, const DAD** subs) {
  int result = 0 ;
  for (int s = 0 ; s < rank ; s++)
    if(!(subs [s])) result++ ;

  return result ;
}

VecGatherSkeleton :: VecGatherSkeleton(const DAD* _loc, const DAD* _rem,
                                       const int len,
                                       const DAD** _subs,
                                       const int** _subsData) :
                         BlockDataSchedule(len,
                                           blockRank(_rem->rnk(), _subs)),
                         loc(_loc), rem(_rem),
                         subs(_subs), subsData(_subsData),
                         rank(_loc->rnk()) {

  Group p = loc->grp() ;
  Group q = rem->grp() ;

  for(int d = 0 ; d < rank ; d++) {
    const DAD* sub = subs [d] ;
    if(sub) {
      if(!p.isEqual(sub->grp()))
	AdlibError("VecGatherSkeleton: Group mismatch") ;
      if(sub->rnk() != 1)
	AdlibError("VecGatherSkeleton: Subscript array must have rank 1") ;
      if(!loc->rng(d).isAligned(sub->rng(0)))
	AdlibError("VecGatherSkeleton: Range misalignment") ;
    }
  }

  if(!apg.contains(p))
    AdlibError("VecGatherSkeleton: loc is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("VecGatherSkeleton: rem is inaccessible.\n") ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc()) {  // Am part of lead reduced group.
      exts     = new int [BlockDataSchedule :: rank] ;
      locSteps = new int [BlockDataSchedule :: rank] ;
      remSteps = new int [BlockDataSchedule :: rank] ;

      loop(0, 1, 0, 1, 1, q, 0, 0) ;

      delete [] remSteps ;
      delete [] locSteps ;
      delete [] exts ;
    }
  }

  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;
  BlockDataSchedule :: build(&spanTree) ;
}

void VecGatherSkeleton :: loop(const int locOffset, const int locCtg,
                               const int remOffset, const int remCtg,
                               const int vol,
                               Group remGrp,
                               const int r, const int s) {

  if(r == rank)
    transReq(locOffset, locSteps, locCtg, remOffset, remSteps, remCtg,
             exts, vol, remGrp.lead_abs()) ;
  else {

    Range x      = loc->rng(r) ;
    Map   locMap = loc->map(r) ;

    Range y      = rem->rng(r) ;
    Map   remMap = rem->map(r) ;

    const DAD* sub = subs [r] ;

    if(sub) {

      // Dimension has a vector subscript.

      Index i(x) ;

      for(i.begin() ; i.test() ; i.next()) {
        const int* subData = subsData [r] ;
        Location j = y(subData [sub->map(0).offset(i)]) ;

	loop(locOffset + locMap.offset(i), locCtg,
             remOffset + remMap.offset(j), remCtg,
             vol,
             remGrp / j, r + 1, s) ;
      }
    }
    else {

      // Dimension is unsubscripted.  (cf `remap').

      LocBlocksIndex i(x) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	int locOffsetElem = locOffset + locMap.offset(i) ;
	const int locStep = locMap.step(i.sub_stp) ;

        AllBlocksIndex j(y.subrng(i.count, i.glb_bas, i.glb_stp)) ;

	for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {

	  const int remOffsetElem = remOffset + remMap.offset(j) ;
	  const int remStep       = remMap.step(j.sub_stp) ;

          const int jlocStep = locStep * j.glb_stp ;

	  exts     [s] = j.count ;
          locSteps [s] = jlocStep ;
	  remSteps [s] = remStep ;

	  loop(locOffsetElem + j.glb_bas * locStep, locCtg && jlocStep == vol,
               remOffsetElem, remCtg && remStep == vol,
               vol * j.count,
               remGrp / j,
               r + 1, s + 1) ;
	}
      }
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherMskSkeleton'                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

VecGatherMskSkeleton :: VecGatherMskSkeleton(const DAD* _loc, const DAD* _rem,
                                             const int len,
                                             const DAD** _subs,
                                             const int** _subsData,
                                             const DAD* _msk,
                                             const int* mskData) :
      DataSchedule(len),
      loc(_loc), rem(_rem), subs(_subs), subsData(_subsData), msk(_msk),
      rank(_loc->rnk()) {

  Group p = loc->grp() ;
  Group q = rem->grp() ;

  for(int d = 0 ; d < rank ; d++) {
    const DAD* sub = subs [d] ;
    if(sub) {
      if(!p.isEqual(sub->grp()))
	AdlibError("VecGatherMskSkeleton: Group mismatch") ;
      if(sub->rnk() != 1)
	AdlibError("VecGatherMskSkeleton: Subscript array must have rank 1") ;
      if(!loc->rng(d).isAligned(sub->rng(0)))
	AdlibError("VecGatherMskSkeleton: Range misalignment") ;
    }
  }
  if(msk)
    alignCheck(loc, msk) ;

  if(!apg.contains(p))
    AdlibError("VecGatherSkeleton: loc is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("VecGatherSkeleton: rem is inaccessible.\n") ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, 0, mskData, q, 0) ;
  }

  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}

void VecGatherMskSkeleton :: loop(const int locOffset,
                                  const int remOffset,
                                  const int* mskElem,
                                  Group remGrp,
                                  const int r) {
  if(r == rank) {
    if(*mskElem)
      transReq(locOffset, remGrp.lead_abs(), remOffset) ;
  }
  else {
    Range x      = loc->rng(r) ;
    Map   locMap = loc->map(r) ;

    Range y      = rem->rng(r) ;
    Map   remMap = rem->map(r) ;

    Map   mskMap = msk->map(r) ;

    const DAD* sub = subs [r] ;
    if(sub) {

      // Dimension has a vector subscript.

      Index i(x) ;

      const int* subData = subsData [r] ;
      Map subMap = sub->map(0) ;

      for(i.begin() ; i.test() ; i.next()) {
        Location j = y(subData [subMap.offset(i)]) ;

	loop(locOffset + locMap.offset(i),
             remOffset + remMap.offset(j),
             mskElem + mskMap.offset(i),
             remGrp / j, r + 1) ;
      }
    }
    else {

      // Dimension is unsubscripted.

      Index i(x) ;

      for(i.begin() ; i.test() ; i.next()) {
//        Location j = y(x.idx(i)) ;
        Location j = y(i) ;

	loop(locOffset + locMap.offset(i),
             remOffset + remMap.offset(j),
             mskElem + mskMap.offset(i),
             remGrp / j, r + 1) ;
      }
    }
  }
}


// -------------------------------------------------------------
// Things to do
//
//   Unify `Reshape' and `ReshapePad' by introducing a start offset in former?
//       Could `reshape' be generalized to support arbitrary F77 ``sequence
//       association'', somehow?
//
//   In any case, `reshape' should be based on a `MessSchedule'...
//       ...it is ``deterministic''.
//
//   Add a version of `VecGatherSkeleton' that works on a `DataSchedule',
//       requiring all dimensions to be vector subscripted.
//
//   Eliminate use of `Index' iterators in in `VecGatherSkeleton',
//       `VecGatherMskSkeleton', and `Reshape' (in favour of lower-level
//        `LocBlocksIndex').
//
//   Where there are several short `int' vectors in a schedule (`exts',
//       `steps', ...) allocate and deallocate them as a single vector.
//       Or allocate them using `Vec'?
//
//   Many calls to `DAD :: rng(const int)' can now be replaced by calls to
//       `Map :: rng()'.
//
//   Delete calls to `restrictToAPG'.
