
// Operations on irregular block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregRangeRep': irregular block-distributed template range.  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

IrregRangeRep :: IrregRangeRep(RangeRep* _kernel, int* _blocks) :
	RangeRep(0, _kernel, _kernel->lev() + 1, _kernel->dim(),
                 DIST_IRREG) {
  kernel->incr() ;

  nBlocks = kernel->size() ;
  map     = new int [nBlocks] ;
  blocks  = new int [nBlocks] ;
  blockSize = 0 ;

  for(int i = 0 ; i < nBlocks ; i++) {
    const int size = _blocks [i] ;
    
    blocks [i] = size ;
    map    [i] = extent ;

    if(size > blockSize) blockSize = size ;
    extent += size ;
  }

  if(extent) scale = ((float) nBlocks) / extent ;
}

IrregRangeRep :: IrregRangeRep(const int _extent, RangeRep* _kernel,
                               Integer* _map) :
	RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim(),
                 DIST_IRREG) {

  // Support for Global Arrays

  kernel->incr() ;

  nBlocks = kernel->size() ;
  map     = new int [nBlocks] ;
  blocks  = new int [nBlocks] ;
  blockSize = 0 ;

  if(extent) scale = ((float) nBlocks) / extent ;

  const int lb = _map [0] ;

  int curr = lb ;
  for(int i = 0 ; i < nBlocks ; i++) {
    const int ip   = i + 1 ;
    const int next = ip < nBlocks ? _map [ip] : lb + extent ;

    if(next < curr)
      AdlibError("IrregRangeRep :: IrregRangeRep: "
                 "_map vector not monotonically increasing\n") ;

    const int size = next - curr ;
    
    blocks [i] = size ;
    map    [i] = curr - lb ;

    if(size > blockSize) blockSize = size ;

    curr = next ;
  }
}

IrregRangeRep :: ~IrregRangeRep() {
  delete [] blocks ;
  delete [] map ;

  kernel->decr() ;
}

RangeRep* IrregRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void IrregRangeRep :: location(Location* loc, const int glb) const {

  loc->crd = findBlock(glb) ;
  loc->sub = glb - map [loc->crd] ;
  loc->dim = dimension ;
}

void IrregRangeRep :: block(Block* blk, const int ker_glb) const {

  // `ker_glb' is a global subscript in the kernel range.

  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = 0 ;
  blk->glb_bas = map [ker_glb] ;

  blk->count   = blocks [ker_glb] ;
}

void IrregRangeRep :: block2(Block* blk, const int ker_glb,
                             const int count, const int base) const {
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  const int blockStart        = map [ker_glb] ;
  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blocks [ker_glb] - 1 ;

  int i_lo, i_hi ;

  i_lo = max(blockStartOffset, 0) ;
  i_hi = min(blockFinishOffset, count - 1) ;

  blk->sub_bas = base + i_lo - blockStart ;
  blk->glb_bas = i_lo ;

  blk->count   = i_hi - i_lo + 1 ;
}

void IrregRangeRep :: block3(Block* blk, const int ker_glb,
                             const int count, const int base,
                             const int stride) const {
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  const int blockStart        = map [ker_glb] ;
  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blocks [ker_glb] - 1 ;

  int i_lo, i_hi ;

  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    i_hi = min(blockFinishOffset / stride, count - 1) ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    i_hi = min(blockStartOffset / stride, count - 1) ;
  }

  blk->sub_bas = base + stride * i_lo - blockStart ;
  blk->glb_bas = i_lo ;

  blk->count   = i_hi - i_lo + 1 ;
}

void IrregRangeRep :: crds2(int* ker_count, int* ker_bas, int* ker_str,
                            const int count, const int base) const {

  int ker_lo = findBlock(base) ;
  int ker_hi = findBlock(base + count - 1) ;

  *ker_bas = ker_lo ;
  *ker_str = 1 ;

  *ker_count = extent ? ker_hi - ker_lo + 1 : 0 ;
}

void IrregRangeRep :: crds3(int* ker_count, int* ker_bas, int* ker_str,
                            const int count, const int base,
                            const int stride) const {

  const int ker_top = findBlock(base + stride * (count - 1)) ;

  *ker_bas = findBlock(base) ;

  if(stride > 0) {
    *ker_count = extent ? ker_top - *ker_bas + 1 : 0 ;
    *ker_str   = 1 ;
  }
  else {
    *ker_count = extent ? *ker_bas - ker_top + 1 : 0 ;
    *ker_str   = -1 ;
  }
}

int IrregRangeRep :: volume() const {
  return blockSize ;
}

int IrregRangeRep :: isAligned(RangeRep* x) const {
  if(x == this) return 1 ;
  if(x->format() == DIST_IRREG && x->dim() == dimension) {
    IrregRangeRep* t = (IrregRangeRep*) x->tem() ;
    for(int i = 0 ; i < nBlocks ; i++)
      if(t->blocks [i] != blocks [i]) return 0 ;
    return x->size() == extent && x->str() == 1 ;
  }
  else
    return 0 ;
}

void IrregRangeRep :: show() {
  AdlibPrintf("IrregRangeRep :: show: size = %d\n", extent) ; 
}

// Obsolescent stuff ...

int IrregRangeRep :: glb(const int sub, const int ker_glb) const {
  return map [ker_glb] + sub ;
}

MEMORY_POOL_DEFN(IrregRangeRep)


// -------------------------------------------------------------
// Things to do
//

