
// Process arrays and groups

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "ProcTree.h"
#include "Procs.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Group'.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void Group :: restrict(Dimension d, const int coord) {
  if(d) { 
    if(!dimSet.member(d))
      AdlibError("Group :: restrict: dimension not active in group") ;

    dimSet  -= d ;
    leadId  += coord * d.str_prc() ;
    isLocal  = isLocal && d.crd() == coord ;
  }
}

void Group :: restrict(Dimension d) {
  if(d) {
    if(!dimSet.member(d))
      AdlibError("Group :: restrict: dimension not active in group") ;

    dimSet -= d ;
    leadId += d.crd() * d.str_prc() ;
  }
}

void Group :: restrict(DimensionSet ds) {
  if(!dims().contains(ds))
    AdlibError("Group :: restrict: "
               "dimension or dimensions not active in group") ;

  for(int r = 0 ; r < procs->rnk() ; r++) {
    Dimension d = procs->dim(r) ;
    if(ds.member(d)) restrict(d) ;
  }
}


int Group :: size() const {

  // Number of processes in the group.

  int size = 1 ;
  for(int r = 0 ; r < procs->rnk() ; r++) {
    Dimension d = procs->dim(r) ;
    if(dimSet.member(d))
      size *= d.size() ;
  }

  return size ;
}

int Group :: id() const {

  // Id of local process relative to this group.

  int index = 0 ;
  for(int r = 0 ; r < procs->rnk() ; r++) {
    Dimension d = procs->dim(r) ;
    if(dimSet.member(d))
      index = d.size() * index + d.crd() ;
  }

  return index ;
}

int Group :: id_prc(const int id) const {

  // Convert id relative to this group to id relative to parent process array.

  int index = id ;
  int parentIndex = leadId ;
  for (int r = procs->rnk() - 1 ; r >= 0 ; r--) {
    Dimension d = procs->dim(r) ;
    if(dimSet.member(d)) {
      int extent = d.size() ;
      parentIndex += d.str_prc() * (index % extent) ;
      index /= extent ;
    }
  }

  return parentIndex ;
}


extern Procs* world ;

int Group :: contains(const Group& p) {
  if(procs == world && dimSet) return 1 ;
      // Quick check for common case: this is the initial process
      // group, therefore contains everything.

  if(procs == p.prc()) {

    // Check if `p' can be obtained from this group by restriction.

    if(dimSet.contains(p.dims())) {
      int index   = leadId ;
      int index_p = p.lead_prc() ;

      for (int r = procs->rnk() - 1 ; r >= 0 ; r--) {
	Dimension d = procs->dim(r) ;
	int extent = d.size() ;

	if(!dimSet.member(d) && index % extent != index_p % extent) return 0 ;

        index   /= extent ;
        index_p /= extent ;
      }

      return 1 ;
    }
    else return 0 ;
  }

  else
    return p.prc() != world && contains(p.prc()->ctx()) ;

        // Check if this group contains the APG that was in effect
        // (the ``context'') when `p.prc()' was created.
}



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Group `apg': the process group executing this control thread.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Group apg ;

void spanTree(ProcTree* tree, const Group& p) {

  // Construct spanning tree of a group of which this process is a member.

  const int index  = p.id() ;
  const int extent = p.size() ;

  tree->numChildren = 0 ;
  for(int i = 1 ; i <= PROCTREE_A ; i++) {
    const int ch = PROCTREE_A * index + i ;
    if(ch < extent) {
      tree->child [i - 1] = p.id_abs(ch) ;
      tree->numChildren++ ;
    }
  }

  if(index == 0)
    tree->isRoot = 1 ;
  else {
    tree->isRoot = 0 ;
    const int pr = (index - 1) / PROCTREE_A ;
    tree->parent = p.id_abs(pr) ;
  }

  tree->root = p.lead_abs() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Procs :: Procs() : rank(1), isLocal(1) {

  // Default constructor for `Procs'.  Create the world---a
  // 1-dimensional process array containing the physical processes.

  // Normally this constructor is invoked as part of the `Adlib'
  // startup sequence---not in user code.

  // `parent' left undefined.

  myId   = AdlibLocId() ;
  mySize = AdlibNProcs() ;

  physIds = new int [mySize] ;
  for(int i = 0 ; i < mySize ; i++)
    physIds [i] = i ;

  dimensions  = new Dimension ;
  dimensions->create(mySize, myId, 1, 0) ;

  dimSet += *dimensions ;
}

static int makeShape(int shp [], int rank, int nprocs) ;

Procs :: Procs(const int _rank, int* _n, int* subIds) :
             rank(_rank) {

  // The standard `Procs' constructor

  // Compute shape and size (`q', `mySize') of array.

  int* q = new int [rank] ;
  if(_rank == 0 || _n) {
    mySize = 1 ;
    for (int d = 0; d < rank ; d++) {
      q [d]   = _n [d] ;
      mySize *= _n [d] ;
    }
    if(mySize > apg.size())
      AdlibError("Procs :: Procs: "
                 "process array too large for active process group\n") ;
  }
  else
    mySize = makeShape(q, rank, apg.size()) ;

  // Initialize `isLocal', `myId', and `physIds' vector.

  isLocal = 0 ;
  myId    = -1 ;
  physIds = new int [mySize] ;

  const int myId_apg = apg.id() ;
  for(int i = 0 ; i < mySize ; i++) {
    int id_apg = subIds ? subIds [i] : i ;
    if(id_apg == myId_apg) {
      isLocal = 1 ;
      myId = i ;
    }
    physIds [i] = apg.id_abs(id_apg) ;
  }

  // Initialize `dimensions'.

//  dimensions = new Dimension [rank] ;
//  int stride = 1 ;
//  for (int r = rank - 1; r >= 0 ; r--) {
//    int coord = isLocal ? (myId / stride) % q [r] : 0 ;
//    dimensions [r].create(q [r], coord, stride, r) ;
//    dimSet += dimensions [r] ;
//    stride *= q [r] ;
//  }

  dimensions = new Dimension [rank] ;
  int stride = 1 ;
  for (int r = 0; r < rank ; r++) {
    int coord = isLocal ? (myId / stride) % q [r] : 0 ;
    dimensions [r].create(q [r], coord, stride, r) ;
    dimSet += dimensions [r] ;
    stride *= q [r] ;
  }

  context = apg ;

  delete [] q ;
}


Procs :: ~Procs() {
  for(int i = 0 ; i < rank ; i++)
    dimensions [i].destroy() ;
  delete [] dimensions ;
  delete [] physIds ;
}


#include <math.h>

static int root(int n, int d) {

  // integer floor of d'th root of `n'.  n >= 0, d > 0.

  switch(d) {
    case 1 :
      return n ;
    case 2 :
      return (int) sqrt(n) ;
    //case 3 :
    //  return (int) cbrt(n) ;
            // (Some systems don't have `cbrt')
    default :
      int ceiling  = 1 ;
      int power    = 1 ;
      while(power < n) {
        ceiling++  ;
        power = ceiling ;
        for(int i = 1 ; i < d ; i++)
          power *= ceiling ;
      }
      if(power == n)
        return ceiling ;
      else
        return ceiling - 1 ;
  }
}

int makeShape(int shp [], int rank, int nprocs) {

  // Compute ``maximal'' shape of specified rank and size <= `nprocs'.

  int size = 1 ;
  while(rank > 0) {
    const int ext = root(nprocs, rank) ;
    rank-- ;
    nprocs /= ext ;
    shp [rank] = ext ;
    size *= ext ;
  }
  return size ;
}


// ---------------------------------------------------------------------
//
// Things to do...
//
//     Make basic functions on `Group' inline.

