
// Schedules for regular remapping of distributed arrays

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "Reduce.h"
#include "Remap.h"


static void shftComponents(int* n, int* exts, int* srcDisps, int* dstDisps,
		           const int extent, const int shift, const Mode mode) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapSkeleton'                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RemapSkeleton :: RemapSkeleton(const DAD* _dst, const DAD* _src,
                               const int len) :
        dst(_dst), src(_src),
        BlockMessSchedule(len, _src->rnk()), rank(_src->rnk()) {

  Group p = src->grp() ;
  Group q = dst->grp() ;

  shapeCheck(dst, src) ;

  if(!apg.contains(p))
    AdlibError("Remap :: Remap: src is inaccessible.\n") ;
  if(!apg.contains(q))
    AdlibError("Remap :: Remap: dst is inaccessible.\n") ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  // Add send requests to queue

  if(p.member()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      sendLoop(0, 1, 1, q, 0) ;
  }

  // Add receive requests to queue

  if(q.member()) {
    DimensionSet rep = q.dims() - dst->sig() ;  // Spans replication group.

    Group reduced = q ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == q.lead_prc())  // Am part of lead reduced group.
      recvLoop(0, 1, 1, p, 0) ;
  }

  delete [] steps ;
  delete [] exts ;

  build() ;
}


void RemapSkeleton :: sendLoop(const int offset, const int ctg, const int vol,
                               Group remGrp, const int r) {
  if(r == rank) {
    if(vol)
      sendReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  }
  else {
    //Range x      = src->rng(r) ;
    Map   locMap = src->map(r) ;

    Range y      = dst->rng(r) ;

    //LocBlocksIndex i(x) ;
    LocBlocksIndex i(locMap.rng()) ;

    // Enumerate locally held blocks of local array range...

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int offsetElem = offset + locMap.offset(i) ;
      const int step = locMap.step(i.sub_stp) ;

      // Local block defines a subrange of local array range.
      // There is a corresponding subrange of the remote array range.
      // Enumerate the blocks in that subrange...

      AllBlocksIndex3 j(y, i.count, i.glb_bas, i.glb_stp) ;

      for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
        const int jstep = step * j.glb_stp ;

	exts  [r] = j.count ;
        steps [r] = jstep ;

	sendLoop(offsetElem + step * j.glb_bas,
                 ctg && jstep == vol, vol * j.count, 
                 remGrp / j,
                 r + 1) ;
      }
    }
  }
}

void RemapSkeleton :: recvLoop(const int offset, const int ctg, const int vol,
                               Group remGrp, const int r) {
  if(r == rank) {
    if(vol)
      recvReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  }
  else {
    //Range x      = dst->rng(r) ;
    Map   locMap = dst->map(r) ;

    Range y      = src->rng(r) ;

    //LocBlocksIndex i(x) ;
    LocBlocksIndex i(locMap.rng()) ;

    // Enumerate locally held blocks of local array range...

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int offsetElem = offset + locMap.offset(i) ;
      const int step = locMap.step(i.sub_stp) ;

      // Local block defines a subrange of local array range.
      // There is a corresponding subrange of the remote array range.
      // Enumerate the blocks in that subrange...

      AllBlocksIndex3 j(y, i.count, i.glb_bas, i.glb_stp) ;

      for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
        const int jstep = step * j.glb_stp ;

	exts  [r] = j.count ;
        steps [r] = jstep ;

	recvLoop(offsetElem + step * j.glb_bas,
                 ctg && jstep == vol, vol * j.count, 
                 remGrp / j,
                 r + 1) ;
      }
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Shift'                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Shift :: Shift(const DAD* _dst, const DAD* _src, const int len,
               const int shift, const int _dim, const Mode mode) :
        BlockMessSchedule(len, _src->rnk()),
	dst(_dst), src(_src), dim(_dim), rank(_src->rnk()) {

  Group p = dst->grp() ;

  if(dim < 0 || dim >= src->rnk())
    AdlibError("Shift :: Shift: dim argument out of range.\n") ;

  alignCheck(dst, src) ;

  if(!apg.contains(p))
    AdlibError("Shift :: Shift: arrays are inaccessible.\n") ;

  Range x = src->rng(dim) ;

  // Set up tables of component contiguous shifts.
  shftComponents(&nShfts, shftExts, shftSrcDisps, shftDstDisps,
		 x.size(), shift, mode) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    // Add send requests to queue

    sendLoop(0, 1, 1, reduced, 0) ;

    // Add receive requests to queue

    recvLoop(0, 1, 1, reduced, 0) ;
  } 

  delete [] steps ;
  delete [] exts ;

  build() ;
}

void Shift :: sendLoop(const int offset, const int ctg, const int vol,
                       Group remGrp, const int r) {
  if(r == rank)
    sendReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = src->rng(r) ;
    Map   locMap = src->map(r) ;
    Range x      = locMap.rng() ;

    if(r == dim)
      for(int i = 0 ; i < nShfts ; i++) {
	const int shftExt     = shftExts     [i] ;
	const int shftLocDisp = shftSrcDisps [i] ;
	const int shftRemDisp = shftDstDisps [i] ;

	LocBlocksIndex2 i(x, shftExt, shftLocDisp) ;

	for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	  int offsetElem = offset + locMap.offset(i) ;
	  const int step = locMap.step(i.sub_stp) ;

          AllBlocksIndex3 j(x, i.count, shftRemDisp + i.glb_bas, i.glb_stp) ;

	  for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
            const int jstep = step * j.glb_stp ;

	    exts  [r] = j.count ;
	    steps [r] = jstep ;

	    sendLoop(offsetElem + j.glb_bas * step,
                     ctg && jstep == vol, vol * j.count,
                     remGrp / j, r + 1) ;
	  }
	}
      }
    else {
      LocBlocksIndex i(x) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
        const int step = locMap.step(i.sub_stp) ;

	steps [r] = step ;
	exts  [r] = i.count ;

	sendLoop(offset + locMap.offset(i), ctg && step == vol, vol * i.count,
                 remGrp, r + 1) ;
      }
    }
  }
}

void Shift :: recvLoop(const int offset, const int ctg, const int vol,
                       Group remGrp, const int r) {
  if(r == rank)
    recvReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = dst->rng(r) ;
    Map   locMap = dst->map(r) ;
    Range x      = locMap.rng() ;

    if(r == dim)
      for(int i = 0 ; i < nShfts ; i++) {
	const int shftExt     = shftExts     [i] ;
	const int shftLocDisp = shftDstDisps [i] ;
	const int shftRemDisp = shftSrcDisps [i] ;

	LocBlocksIndex2 i(x, shftExt, shftLocDisp) ;

	for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	  int offsetElem = offset + locMap.offset(i) ;
	  const int step = locMap.step(i.sub_stp) ;

          AllBlocksIndex3 j(x, i.count, shftRemDisp + i.glb_bas, i.glb_stp) ;

	  for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
            const int jstep = step * j.glb_stp ;

	    exts  [r] = j.count ;
	    steps [r] = jstep ;

	    recvLoop(offsetElem + j.glb_bas * step,
                    ctg && jstep == vol, vol * j.count,
                    remGrp / j, r + 1) ;
	  }
	}
      }
    else {
      LocBlocksIndex i(x) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
        const int step = locMap.step(i.sub_stp) ;

	steps [r] = step ;
	exts  [r] = i.count ;

	recvLoop(offset + locMap.offset(i), ctg && step == vol, vol * i.count,
                 remGrp, r + 1) ;
      }
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Class `Skew'                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Skew:: Skew(const DAD* _dst, const DAD* _src, const int len,
            const DAD* _shf, int* shfDat,
            const int _dim, const Mode _mode) :
        BlockMessSchedule(len, 1),
	dst(_dst), src(_src), shf(_shf), dim(_dim), mode(_mode) {

  Group p = dst->grp() ;

  if(dim < 0 || dim >= src->rnk())
    AdlibError("Skew :: Skew: dim argument out of range.\n") ;

  alignCheck(dst, src) ;
  alignCheck(src, shf, dim) ;

  if(!apg.contains(p))
    AdlibError("Skew :: Skew: arrays are inaccessible.\n") ;

  rank = src->rnk() ;

  if(p.member()) {
    Range x = src->rng(dim) ;
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    loop(0, 0, shfDat, 0, 0) ;
  }

  build() ;
}

void Skew :: loop(const int dstOffset, const int srcOffset,
                  int* shfData, const int r, const int s) {

  if(r == rank) {
    //Range x    = src->rng(dim) ;

    Map dstMap = dst->map(dim) ;
    Map srcMap = src->map(dim) ;

    Range x    = srcMap.rng() ;

    int nShfts ;
    int shftExts [2] ;
    int shftSrcDisps [2] ;
    int shftDstDisps [2] ;
    shftComponents(&nShfts, shftExts, shftSrcDisps, shftDstDisps,
		   x.size(), *shfData, mode) ;

    for(int i = 0 ; i < nShfts ; i++) {
      const int shftExt     = shftExts     [i] ;
      const int shftSrcDisp = shftSrcDisps [i] ;
      const int shftDstDisp = shftDstDisps [i] ;

      shiftSend(srcOffset, srcMap, shftExt, shftSrcDisp, shftDstDisp) ;
      shiftRecv(dstOffset, dstMap, shftExt, shftDstDisp, shftSrcDisp) ;
    }
  }
  else if(r == dim)
    loop(dstOffset, srcOffset, shfData, r + 1, s) ;
  else {
    //Range x      = src->rng(r) ;

    Map   dstMap = dst->map(r) ;
    Map   srcMap = src->map(r) ;

    Range x    = srcMap.rng() ;

    Map   shfMap = shf->map(s) ;

    LocBlocksIndex i(x) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      int dstOffsetElem = dstOffset + dstMap.offset(i) ;
      int dstStep = dstMap.step(i.sub_stp) ;

      int srcOffsetElem = srcOffset + srcMap.offset(i) ;
      int srcStep = srcMap.step(i.sub_stp) ;

      int* shfDataElem = shfData + shfMap.offset(i) ;
      int shfStep = shfMap.step(i.sub_stp) ;

      for(int sub = 0 ; sub < i.count ; sub++) {
	loop(dstOffsetElem, srcOffsetElem, shfDataElem, r + 1, s + 1) ;

        dstOffsetElem += dstStep ;
        srcOffsetElem += srcStep ;
        shfDataElem += shfStep ;
      }
    }
  }
}

void Skew :: shiftSend(int offset, Map map,
                       const int ext, const int locDisp, const int remDisp) {

  //Range x = src->rng(dim) ;

  Range x = map.rng() ;

  LocBlocksIndex2 i(x, ext, locDisp) ;
  for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

    int offsetElem = offset + map.offset(i) ;
    int step = map.step(i.sub_stp) ;

    AllBlocksIndex3 j(x, i.count, remDisp + i.glb_bas, i.glb_stp) ;

    for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
      Group p = reduced / j ; 

      int s = step * j.glb_stp;
      
      sendReq(offsetElem + j.glb_bas * step, &s,s == 1,
              &j.count, j.count, p.lead_abs()) ;
    }
  }
}

void Skew :: shiftRecv(int offset, Map map,
                       const int ext, const int locDisp, const int remDisp) {

  Range x = src->rng(dim) ;

  LocBlocksIndex i(x.subrng(ext, locDisp)) ;
  for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

    int offsetElem = offset + map.offset(i) ;
    int step = map.step(i.sub_stp) ;

    AllBlocksIndex3 j(x, i.count, remDisp + i.glb_bas, i.glb_stp) ;

    for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
      Group p = reduced / j ; 

      int s = step * j.glb_stp;
      
      recvReq(offsetElem + j.glb_bas * step, &s,s == 1,
              &j.count, j.count, p.lead_abs()) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `MultiShift'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

MultiShift :: MultiShift(const DAD* _dst, const DAD* _src, const int len,
                         int* shift, const Mode* mode) :
        BlockMessSchedule(len, _src->rnk()),
	dst(_dst), src(_src), rank(_src->rnk()) {

  Group p = dst->grp() ;

  alignCheck(dst, src) ;

  if(!apg.contains(p))
    AdlibError("MultiShift :: MultiShift: arrays are inaccessible.\n") ;

  nShfts       = new int [rank] ;
  shftExts     = new int [rank] [2] ;
  shftSrcDisps = new int [rank] [2] ;
  shftDstDisps = new int [rank] [2] ;

  // Set up tables of component contiguous shifts.
  for(int r = 0 ; r < rank ; r++)
    shftComponents(&(nShfts [r]),
                   shftExts [r], shftSrcDisps [r], shftDstDisps [r],
		   src->rng(r).size(), shift [r], mode [r]) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  if(p.member()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    // Add send requests to queue

    sendLoop(0, 1, 1, reduced, 0) ;

    // Add receive requests to queue

    recvLoop(0, 1, 1, reduced, 0) ;
  }

  delete [] steps ;
  delete [] exts ;

  delete [] nShfts ;
  delete [] shftExts ;
  delete [] shftSrcDisps ;
  delete [] shftDstDisps ;

  build() ;
}

void MultiShift :: sendLoop(const int offset, const int ctg, const int vol,
                            Group remGrp, const int r) {
  if(r == rank)
    sendReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = src->rng(r) ;
    Map   locMap = src->map(r) ;
    Range x      = locMap.rng() ;

    for(int shf = 0 ; shf < nShfts [r] ; shf++) {
      const int shftExt     = shftExts     [r] [shf] ;
      const int shftLocDisp = shftSrcDisps [r] [shf] ;
      const int shftRemDisp = shftDstDisps [r] [shf] ;

      LocBlocksIndex2 i(x, shftExt, shftLocDisp) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	int offsetElem = offset + locMap.offset(i) ;
	const int step = locMap.step(i.sub_stp) ;

	steps [r] = step ;

        AllBlocksIndex2 j(x, i.count, shftRemDisp + i.glb_bas) ;

	for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
	  const int count = j.count ;

	  exts [r] = count ;

	  sendLoop(offsetElem, ctg && step == vol, vol * count,
                   remGrp / j, r + 1) ;

	  offsetElem += count * step ;
	}
      }
    }
  }
}

void MultiShift :: recvLoop(const int offset, const int ctg, const int vol,
                            Group remGrp, const int r) {
  if(r == rank)
    recvReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = dst->rng(r) ;
    Map   locMap = dst->map(r) ;
    Range x      = locMap.rng() ;

    for(int shf = 0 ; shf < nShfts [r] ; shf++) {
      const int shftExt     = shftExts     [r] [shf] ;
      const int shftLocDisp = shftDstDisps [r] [shf] ;
      const int shftRemDisp = shftSrcDisps [r] [shf] ;

      LocBlocksIndex2 i(x, shftExt, shftLocDisp) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	int offsetElem = offset + locMap.offset(i) ;
	const int step = locMap.step(i.sub_stp) ;

	steps [r] = step ;

        AllBlocksIndex2 j(x, i.count, shftRemDisp + i.glb_bas) ;

	for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
	  const int count = j.count ;

	  exts [r] = count ;

	  recvLoop(offsetElem, ctg && step == vol, vol * count,
                   remGrp / j, r + 1) ;

	  offsetElem += count * step ;
	}
      }
    }
  }
}


static void shftComponents(int* n, int* exts, int* srcDisps, int* dstDisps,
		           const int extent, const int shift, const Mode mode) {
  *n = 0 ;

  switch(mode) {
    case CYCL : {
      const int shiftMod =
	  shift < 0 ? (shift % extent + extent) : (shift % extent) ;

      exts     [*n] = extent - shiftMod ;
      srcDisps [*n] = shiftMod ;
      dstDisps [*n] = 0 ;
      (*n)++ ;

      if(shiftMod != 0) {
	exts     [*n] = shiftMod ;
	srcDisps [*n] = 0 ;
	dstDisps [*n] = extent - shiftMod ;
	(*n)++ ;
      }
                  

      break ;
    }
    case EDGE : {
      if(shift >= 0 && shift < extent) {
	exts     [*n] = extent - shift ;
	srcDisps [*n] = shift ;
	dstDisps [*n] = 0 ;
	(*n)++ ;
      }
      
      if(shift < 0 && shift > -extent) {
	exts     [*n] = extent + shift ;
	srcDisps [*n] = 0 ;
	dstDisps [*n] = -shift ;
	(*n)++ ;
      }

      break ;
    }
    case NONE :
      break ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `WriteFaces'.  Write upper and lower ghost areas for          //
//                      dimension `dim'.                               //
//                                                                     //
//                      Copied area includes ghost extensions          //
//                      associated with dimensions r < dim.  (But not  //
//                      those associated with dimensions r > dim.)     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline int imax(int x, int y) {return x > y ? x : y ; }
inline int imin(int x, int y) {return x < y ? x : y ; }

WriteFaces :: WriteFaces(const DAD* _src, const int len, const int _dim,
                         const int* _wlo, const int* _whi, const Mode* _mode) :
        BlockMessSchedule(len, _src->rnk()),
        src(_src), dim(_dim), wlo(_wlo), whi(_whi), mode(_mode),
        rank(_src->rnk()) {

  Range x = src->rng(dim) ;
  Group p = src->grp() ;

  if(!apg.contains(p))
    AdlibError("WriteFaces :: WriteFaces: array is inaccessible.\n") ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    loop(0, 1, 1, reduced, 0) ;
  }

  delete [] steps ;
  delete [] exts ;

  build() ;
}

void WriteFaces :: loop(const int offset,
                        const int ctg, const int vol,
		        const Group& remGrp, const int r) {

  //Range x      = src->rng(r) ;
  Map   locMap = src->map(r) ;
  Range x      = locMap.rng() ;

  LocBlocksIndex i(x) ;
  const int N = x.size() ;

  if(r == dim) {
    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      int offsetElem = offset + locMap.offset(i) ;
      const int step = locMap.step(i.sub_stp) ;

      switch(mode ? mode [r] : EDGE) {
        case EDGE : {

          // (Probably should absorb this case into the implementation of CYCL,
          //  to avoid some repetition).

	  if(wlo [r] > 0) {

	    // Sending data to lower ghost edges.

	    // Iterate over a patch of width `wlor' to the right of the
	    // current block.  This includes start points of all blocks whose
	    // lower ghost area includes an image of part of (or all of) the
	    // current block.  (In the most typical case there will only be
	    // one such block, but in general there is no guarantee that the
	    // adjacent blocks are wider than the ghost area itself).

	    const int neighbBase = i.glb_bas + i.count ;
	    const int wlor = imin(wlo [r], N - neighbBase) ;
	    AllBlocksIndex2 jlor(x, wlor, neighbBase) ;

	    for(jlor.beginAllBlk() ; jlor.test() ; jlor.nextAllBlk()) {

	      // Send overlap of current block (`i') with lower ghost region
	      // of block `jlor' to the owner of block `jlor'.

	      const int start = imax(i.count + jlor.glb_bas - wlo [r], 0) ;
	      const int count = i.count - start ;

	      exts  [r] = count ;
	      steps [r] = step ;

	      sendLoop(offsetElem + start * step,
		       ctg && step == vol, vol * count,
		       remGrp / jlor, r + 1) ;
	    }

	    // Receiving data for lower ghost edges.

	    // Iterate over a patch of width `wlol' to the left of the
	    // current block.

	    const int wlol = imin(wlo [r], i.glb_bas) ;
	    AllBlocksIndex2 jlol(x, wlol, i.glb_bas - wlol) ;

	    for(jlol.beginAllBlk() ; jlol.test() ; jlol.nextAllBlk()) {
	      exts  [r] = jlol.count ;
	      steps [r] = step ;

	      recvLoop(offsetElem + (jlol.glb_bas - wlol) * step,
		       ctg && step == vol, vol * jlol.count,
		       remGrp / jlol, r + 1) ;
	    }
	  }

	  if(whi [r] > 0) {

	    // Sending data to upper ghost edges.

	    // Iterate over a patch of width `whil' to the left of the
	    // current block.  This includes end points of all blocks whose
	    // upper ghost area is covered by the current block.

	    const int whil = imin(whi [r], i.glb_bas) ;
	    AllBlocksIndex2 jhil(x, whil, i.glb_bas - whil) ;

	    for(jhil.beginAllBlk() ; jhil.test() ; jhil.nextAllBlk()) {

	      // Send overlap of current block (`i') with lower ghost region
	      // of block `jhil' to the owner of block `jhil'.

	      const int count = imin(whi [r] -
                                    (whil - jhil.glb_bas - jhil.count),
				    i.count) ;

	      exts  [r] = count ;
	      steps [r] = step ;

	      sendLoop(offsetElem,
		       ctg && step == vol, vol * count,
		       remGrp / jhil, r + 1) ;
	    }

	    // Receiving data for upper ghost edges.

	    // Iterate over a patch of width `whir' to the right of the
	    // current block.

	    const int neighbBase = i.glb_bas + i.count ;
	    const int whir = imin(whi [r], N - neighbBase) ;
	    AllBlocksIndex2 jhir(x, whir, neighbBase) ;

	    for(jhir.beginAllBlk() ; jhir.test() ; jhir.nextAllBlk()) {
	      exts  [r] = jhir.count ;
	      steps [r] = step ;

	      recvLoop(offsetElem + (i.count + jhir.glb_bas) * step,
		       ctg && step == vol, vol * jhir.count,
		       remGrp / jhir, r + 1) ;
	    }
	  }
          break ;
        }
	case CYCL : {

	  int subrng_base, residue ;

	  if(wlo [r] > 0) {

	    // Sending data to lower ghost edges.

	    // Iterate over a patch of width `wlo [r]' to the right of
	    // the current block.  This patch includes start points
	    // of all blocks whose lower ghost area includes an image
	    // of part of (or all of) the current block.  The patch may
	    // wrap past the end of the range of the array, and thus be
	    // split into two subranges.  Actually, we have to take
	    // account of the possibility that it could split into
	    // *more* than two subranges, because the patch could wrap
	    // round more than once---in general there is no guarantee
	    // that the ghost region width is smaller than the entire
	    // range of the array!

	    // Parameters for first subrange in patch:

	    subrng_base = (i.glb_bas + i.count) % N ;
	    residue     = wlo [r] ;

	    // Loop over contiguous subranges of the ``patch''...

	    while(residue) {
	      const int subrng_size = imin(residue, N - subrng_base) ;

	      AllBlocksIndex2 jlor(x, subrng_size, subrng_base) ;

	      for(jlor.beginAllBlk() ; jlor.test() ; jlor.nextAllBlk()) {

		// Send overlap of current block, `i', with lower ghost
		// region of block `jlor' to the owner of block `jlor'.

		const int residueJ = residue - jlor.glb_bas ;

		  // Residue of patch after start of destination block
		  // `jlor'---this is an upper bound for the overlap size:
                  //
                  //                                 residueJ
                  //                             |<--+-->|
                  //               |<--- wlo (patch) --->|
                  //
                  //   |           |             |           |
                  //   |     i     |             |    jlor   |
                  //   |           |             |           |
                  //
                  //       |<--- wlo (ghost) --->|
                  //       |<--+-->|
                  //           overlap
                  //   |<+>|
                  //     start

		const int overlap = imin(residueJ, i.count) ;
		    // Size of overlap.

		const int start = i.count - overlap ;
		    // Start displacement of overlap in current block.

		exts  [r] = overlap ;
		steps [r] = step ;

		sendLoop(offsetElem + start * step,
			 ctg && step == vol, vol * overlap,
			 remGrp / jlor, r + 1) ;
	      }

	      // Parameters for next subrange in patch:

	      subrng_base  = 0 ;  // Wrapped to start of range.
	      residue     -= subrng_size ;
	    }

	    // Receiving data for lower ghost edges.

	    // Iterate over a patch of width `wlo [r]' to the left of the
	    // current block.

	    // Parameters for first subrange in patch:

	    subrng_base = N - (N - i.glb_bas + wlo [r]) % N ;
	    residue = wlo [r] ;

	    // Loop over contiguous subranges of the ``patch''...

	    while(residue) {
	      const int subrng_size = imin(residue, N - subrng_base) ;

	      AllBlocksIndex2 jlol(x, subrng_size, subrng_base) ;

	      for(jlol.beginAllBlk() ; jlol.test() ; jlol.nextAllBlk()) {

		const int residueJ = residue - jlol.glb_bas ;

		  // Residue of patch after start of source block
                  // `jlol'---this subtracts from the destination offset.
                  //
                  //   |<------- redidueJ ------>|
                  //
                  //   |           |             |           |
                  //   |    jlol   |             |     i     |
                  //   |           |             |           |
                  //

		exts  [r] = jlol.count ;
		steps [r] = step ;

		recvLoop(offsetElem - residueJ * step,
			 ctg && step == vol, vol * jlol.count,
			 remGrp / jlol, r + 1) ;
	      }

	      // Parameters for next subrange in patch:

	      subrng_base  = 0 ;  // Wrapped to start of range.
	      residue     -= subrng_size ;
	    }
	  }

	  if(whi [r] > 0) {

	    // Sending data to upper ghost edges.

	    // (If ghost regions symmetric---a common case---can merge this
            //  with loop for receiving data for lower ghost edges).

	    // Iterate over a patch of width `whi [r]' to the left of
	    // the current block.  This includes end points of all
	    // blocks whose upper ghost area is covered by the current
	    // block.  The ``patch'' may wrap past the start of the
	    // range of the array, and thus be split into multiple
	    // subranges.

	    // Parameters for first subrange in patch:

	    subrng_base = N - (N - i.glb_bas + whi [r]) % N ;
	    residue = whi [r] ;

	    // Loop over contiguous subranges of the ``patch''...

	    while(residue) {
	      const int subrng_size = imin(residue, N - subrng_base) ;

	      AllBlocksIndex2 jhil(x, subrng_size, subrng_base) ;

	      for(jhil.beginAllBlk() ; jhil.test() ; jhil.nextAllBlk()) {

		// Send overlap of current block, `i', with lower ghost
		// region of block `jhil' to the owner of block `jhil'.

		const int residueJ = residue - jhil.glb_bas ;

		  // Residue of patch after start of destination block
		  // `jhil'---this subtracts from the overlap size.
                  //
                  //     jhil.count
                  //    |<----+---->|<--- whi (ghost) --->|
                  //
                  //    |<------ residueJ ------->|
                  //
                  //    |           |             |           |
                  //    |    jhil   |             |     i     |
                  //    |           |             |           |
                  //
                  //                              |<--+-->|
                  //                                  overlap

		const int overlap = imin(jhil.count + whi [r] - residueJ,
                                        i.count) ;  // Size of overlap.

		exts  [r] = overlap ;
		steps [r] = step ;

		sendLoop(offsetElem,
			 ctg && step == vol, vol * overlap,
			 remGrp / jhil, r + 1) ;
	      }

	      // Parameters for next subrange in patch:

	      subrng_base  = 0 ;  // Wrapped to start of range.
	      residue     -= subrng_size ;
	    }

	    // Receiving data for upper ghost edges.

	    // Iterate over a patch of width `whi [r]' to the right of the
	    // current block.

	    // (If ghost regions symmetric---a common case---can merge this
            //  with loop for sending data to lower ghost edges).

	    // Parameters for first subrange in patch:

	    subrng_base = (i.glb_bas + i.count) % N ;
	    residue     = whi [r] ;

	    // Loop over contiguous subranges of the ``patch''...

	    while(residue) {
	      const int subrng_size = imin(residue, N - subrng_base) ;

	      AllBlocksIndex2 jhir(x, subrng_size, subrng_base) ;

	      for(jhir.beginAllBlk() ; jhir.test() ; jhir.nextAllBlk()) {

		const int residueJ = residue - jhir.glb_bas ;

		  // Residue of patch after start of source block
                  // `jhir'---this subtracts from the destination offset.
                  //
                  //                                 residueJ
                  //                             |<--+-->|
                  //      i.count
                  //   |<----+---->|<--- whi (ghost) --->|
                  //
                  //   |           |             |           |
                  //   |     i     |             |    jhir   |
                  //   |           |             |           |
                  //
                  //   |<----- offset shift ---->|

		exts  [r] = jhir.count ;
		steps [r] = step ;

		recvLoop(offsetElem + (i.count + whi [r] - residueJ) * step,
			 ctg && step == vol, vol * jhir.count,
			 remGrp / jhir, r + 1) ;
	      }

	      // Parameters for next subrange in patch:

	      subrng_base  = 0 ;  // Wrapped to start of range.
	      residue     -= subrng_size ;
	    }
	  }
          break ;
	}
        case NONE :
          break ;
      }
    }
  }
  else {

    // r < dim.  Include ghost extensions for orthogonal dimension in copy.

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      const int step  = locMap.step(i.sub_stp) ;
      const int count = i.count + wlo [r] + whi [r] ;

      steps [r] = step ;
      exts  [r] = count ;

      // offset + locMap.offset(i - wlo [r])
      loop(offset + locMap.offset(i) - wlo [r] * step,
	   ctg && step == vol, vol * count,
	   remGrp, r + 1) ;
    }
  }
}

void WriteFaces :: recvLoop(const int offset,
                            const int ctg, const int vol,
                            const Group& remGrp, const int r) {
  if(r == rank)
    recvReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = src->rng(r) ;
    Map   locMap = src->map(r) ;
    Range x      = locMap.rng() ;

    LocBlocksIndex i(x) ;
    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      const int step = locMap.step(i.sub_stp) ;

      steps [r] = step ;
      exts  [r] = i.count ;

      recvLoop(offset + locMap.offset(i),
	       ctg && step == vol, vol * i.count,
	       remGrp, r + 1) ;
    }
  }
}

void WriteFaces :: sendLoop(const int offset,
                            const int ctg, const int vol,
                            const Group& remGrp, const int r) {
  if(r == rank)
    sendReq(offset, steps, ctg, exts, vol, remGrp.lead_abs()) ;
  else {
    //Range x      = src->rng(r) ;
    Map   locMap = src->map(r) ;
    Range x      = locMap.rng() ;

    LocBlocksIndex i(x) ;
    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      const int step = locMap.step(i.sub_stp) ;

      steps [r] = step ;
      exts  [r] = i.count ;

      sendLoop(offset + locMap.offset(i),
	       ctg && step == vol, vol * i.count,
	       remGrp, r + 1) ;
    }
  }
}

// -------------------------------------------------------------
// Things to do
//
//   Where there are several short `int' vectors in a schedule (`exts',
//       `steps', ...) allocate and deallocate them as a single vector.
//       Or allocate them using `Vec'?
//
//   Contiguity checks should test for count = 1, eg:
//       ctg && (i.count == 1 || step == vol)
//       ?
//
//   Many calls to `DAD :: rng(const int)' can now be replaced by calls to
//       `Map :: rng()'.
//
//   In `WriteFaces', merge some loops if ghost areas are symmetric.
//       Merge implementation of `EDGE' case with `CYCL' case.
//
//   Optional run-time checks should include check that written
//       ghost areas fit in those of array.
//
//   It would be convenient if `WriteHalo' defaulted to update
//       the entire halo area of an array.

