
// Schedules for tree-based communications.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "threads.h"
#include "adtags.h"
#include "ProcTree.h"
#include "schedule.h"
#include "TreeSchedule.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `TreeSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void TreeSchedule :: synch() {
  int i ;
  MPI_Status status ;

  for(i = 0 ; i < numChildren ; i++) {
    MPI_Irecv(NULL, 0, MPI_INT,
              child [i], BAR, MPI_COMM_WORLD, event) ;

    Server :: waitClientEvent(&status) ;
  }

  if(!isRoot) {
    MPI_Isend(NULL, 0, MPI_INT,
              parent, BAR, MPI_COMM_WORLD, event) ;

    Server :: waitClientEvent(&status) ;

    MPI_Irecv(NULL, 0, MPI_INT,
              parent, RAB, MPI_COMM_WORLD, event) ;

    Server :: waitClientEvent(&status) ;
  }

  for(i = numChildren - 1 ; i >= 0 ; i--) {
    MPI_Isend(NULL, 0, MPI_INT,
              child [i], RAB, MPI_COMM_WORLD, event) ;

    Server :: waitClientEvent(&status) ;
  }
}

void TreeSchedule :: broadcast(const int resLen, void* res) {
  MPI_Status status ;

  if(!isRoot)
    MPI_Recv((char*) res, resLen, MPI_CHAR,
             parent, DAT, MPI_COMM_WORLD, &status) ;

  int i ;

  for(i = 0 ; i < numChildren ; i++)
    MPI_Send((char*) res, resLen, MPI_CHAR, child [i], DAT, MPI_COMM_WORLD) ;
}

void TreeSchedule :: broadcast(const int resLen, void* res, const int srcId) {
  if(srcId == root)
    broadcast(resLen, res) ;
  else {
    // Logically, perform broadcast from `srcId' on the new tree obtained by:
    //
    //   On the source process: Add `root' as a new child (the first child).
    //   On the root process:   Add `srcId' as my parent.
    //   On any process:        Delete `srcId' as a child.
    //
    // ie, delete the incoming arc to `srcId', and add an outgoing arc to
    // `root'.

    int locId ;
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;

    MPI_Status status ;

    if(locId != srcId)
      MPI_Recv((char*) res, resLen, MPI_CHAR,
	       isRoot ? srcId : parent, DAT, MPI_COMM_WORLD, &status) ;
    else
      MPI_Send((char*) res, resLen, MPI_CHAR, root, DAT, MPI_COMM_WORLD) ;

    for(int i = 0 ; i < numChildren ; i++)
      if(child [i] != srcId)
	MPI_Send((char*) res, resLen, MPI_CHAR,
		 child [i], DAT, MPI_COMM_WORLD) ;
  }
}

void TreeSchedule :: broadcast(const int res1Len, void* res1,
                               const int res2Len, void* res2, const int srcId) {
  int resLen = res1Len + res2Len ;
  if(resLen <= STATIC_BUFFER_SIZE) {
    int locId ;
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;

    if(locId == srcId) {
      memcpy((char*) statBuf, (char*) res1, res1Len) ;
      memcpy((char*) statBuf + res1Len, (char*) res2, res2Len) ;
      broadcast(resLen, statBuf, srcId) ;
    }
    else {
      broadcast(resLen, statBuf, srcId) ;
      memcpy((char*) res1, (char*) statBuf,           res1Len) ;
      memcpy((char*) res2, (char*) statBuf + res1Len, res2Len) ;
    }
  }
  else {
    broadcast(res1Len, res1, srcId) ;
    broadcast(res2Len, res2, srcId) ;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RedxSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void RedxSchedule :: combine(const int resLen, void* res) {
  MPI_Status status ;

  char* buffer ;
  if(resLen > STATIC_BUFFER_SIZE)
    buffer = new char [resLen] ;  // *** potential word-alignment problem ***
  else
    buffer = (char*) statBuf ;

  for(int i = 0 ; i < numChildren ; i++) {
    MPI_Recv(buffer, resLen, MPI_CHAR, child [i], DAT,
             MPI_COMM_WORLD, &status) ;

    opCombine(buffer) ;
  }

  if(resLen > STATIC_BUFFER_SIZE)
    delete [] buffer ;

  if(!isRoot)
    MPI_Send(res, resLen, MPI_CHAR, parent, DAT, MPI_COMM_WORLD) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `Redx2Schedule' class.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void Redx2Schedule :: combine(const int res1Len, void* res1,
                              const int res2Len, void* res2) {
  MPI_Status status ;

  const int a_res1Len = ALIGN_UP(res1Len) ;
  const int a_res2Len = ALIGN_UP(res2Len) ;
  int resLen = a_res1Len + a_res2Len ;

  char* buffer ;
  if(resLen > STATIC_BUFFER_SIZE)
    buffer = new char [resLen] ;  // *** potential word-alignment problem ***
  else
    buffer = (char*) statBuf ;

  for(int i = 0 ; i < numChildren ; i++) {
    MPI_Recv(buffer, resLen, MPI_CHAR, child [i], DAT,
             MPI_COMM_WORLD, &status) ;

    opCombine(buffer, buffer + a_res1Len) ;
  }

  if(!isRoot) {
    memcpy(buffer,             (char*) res1, res1Len) ;
    memcpy(buffer + a_res1Len, (char*) res2, res2Len) ;
    MPI_Send(buffer, resLen, MPI_CHAR, parent, DAT, MPI_COMM_WORLD) ;
  }

  if(resLen > STATIC_BUFFER_SIZE)
    delete [] buffer ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockTreeSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockTreeSchedule :: broadcast(void* base) {
  MPI_Status status ;

  Mess* message = messages.data() ;

  for(int mess = 0 ; mess < messages.size() ; mess++) {
    int size = message->size ;
    int num  = message->num  ;
    MultiBlock block = blocks.data() ;

    if(num == 1 && block [MBLK_CTG]) {

      // Message is single contiguous block.  Send/receive direct from/to
      // array block, wherever possible.

      char* blockBase  = (char*) base + elementLen * block [MBLK_OFF] ;

      if(!isRoot)
	// Receive block of array from parent
	MPI_Recv(blockBase, size, MPI_CHAR,
		 parent, DAR, MPI_COMM_WORLD, &status) ;

      // Send block of array to children
      for(int i = 0 ; i < numChildren ; i++)
	MPI_Send(blockBase, size, MPI_CHAR, child [i], DAR, MPI_COMM_WORLD) ;

      nextBlock(block) ;
    }
    else {

      if(isRoot) {
	// Copy blocks of array to buffer
	char* ptr = (char*) statBuf ;
	for(int j = 0 ; j < num ; j++) {
	  gatherMblkBuf(base, block, ptr, elementLen) ;
	  ptr += block [MBLK_SIZ] ;
	  nextBlock(block) ;
	}
      }
      else {
        // Receive buffer from parent
	MPI_Recv(statBuf, size, MPI_CHAR,
                 parent, DAR, MPI_COMM_WORLD, &status) ;

        // Copy buffer to blocks of array
	char* ptr = (char*) statBuf ;
	for(int j = 0 ; j < num ; j++) {
	  scatterMblkBuf(base, block, ptr, elementLen) ;
	  ptr += block [MBLK_SIZ] ;
	  nextBlock(block) ;
	}
      }

      // Send buffer to children
      for(int i = 0 ; i < numChildren ; i++)
	MPI_Send(statBuf, size, MPI_CHAR, child [i], DAR, MPI_COMM_WORLD) ;
    }

    message++ ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRedxSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockRedxSchedule :: combine(void* base) {
  MPI_Status status ;

  Mess*    message = messages.data() ;
  MultiBlock block = blocks.data() ;

  for(int mess = 0 ; mess < messages.size() ; mess++) {
    int size    = message->size ;
    int num     = message->num  ;

    char* buffer ;
    if(size > STATIC_BUFFER_SIZE)
          // Only really expected if elementLen > STATIC_BUFFER_SIZE.  Rare??
      buffer = new char [size] ;  // *** potential word-alignment problem ***
    else
      buffer = (char*) statBuf ;

    if(num == 1 && block [MBLK_CTG]) {

      // Message is single contiguous block.  Send/receive direct from/to
      // array block, wherever possible.

      char* blockBase  = (char*) base + elementLen * block [MBLK_OFF] ;
      int i ;

      // Receive buffer from children, and combine with blocks of array.
      for(i = 0 ; i < numChildren ; i++) {
	MPI_Recv(buffer, size, MPI_CHAR,
	         child [i], DAR, MPI_COMM_WORLD, &status) ;

	opCombine(block, buffer) ;
      }

      if(!isRoot) {

	// Send block of array to parent
	MPI_Send(blockBase, size, MPI_CHAR, parent, DAR, MPI_COMM_WORLD) ;

	// Receive block of array from parent
	MPI_Recv(blockBase, size, MPI_CHAR,
		 parent, DAR, MPI_COMM_WORLD, &status) ;
      }

      // Send block of array to children
      for(i = 0 ; i < numChildren ; i++) {
	MPI_Send(blockBase, size, MPI_CHAR, child [i], DAR, MPI_COMM_WORLD) ;
      }

      nextBlock(block) ;
    }
    else {
      MultiBlock firstBlock = block ;
      int i ;

      // Receive buffer from children, and combine with blocks of array.
      for(i = 0 ; i < numChildren ; i++) {

	MPI_Recv(buffer, size, MPI_CHAR,
		 child [i], DAR, MPI_COMM_WORLD, &status) ;

	MultiBlock chBlock = firstBlock ;
	char* ptr = buffer ;
	for(int j = 0 ; j < num ; j++) {
	  opCombine(chBlock, ptr) ;
	  ptr += chBlock [MBLK_SIZ] ;
	  nextBlock(chBlock) ;
	}
      }

      // Copy blocks of array to buffer
      char* ptr = buffer ;
      for(int j = 0 ; j < num ; j++) {
        gatherMblkBuf(base, block, ptr, elementLen) ;
        ptr += block [MBLK_SIZ] ;
	nextBlock(block) ;
      }

      if(!isRoot) {
        // Send buffer to parent
	MPI_Send(buffer, size, MPI_CHAR, parent, DAR, MPI_COMM_WORLD) ;

        // Receive buffer from parent
	MPI_Recv(buffer, size, MPI_CHAR, parent, DAR, MPI_COMM_WORLD, &status) ;

        // Copy buffer to blocks of array
        MultiBlock chBlock = firstBlock ;
	char* ptr = buffer ;
	for(int j = 0 ; j < num ; j++) {
	  scatterMblkBuf(base, chBlock, ptr, elementLen) ;
	  ptr += chBlock [MBLK_SIZ] ;
	  nextBlock(chBlock) ;
	}
      }

      // Send buffer to children
      for(i = 0 ; i < numChildren ; i++)
	MPI_Send(buffer, size, MPI_CHAR, child [i], DAR, MPI_COMM_WORLD) ;
    }

    if(size > STATIC_BUFFER_SIZE)
      delete buffer ;

    message++ ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRedx2Schedule' class.                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockRedx2Schedule :: combine(void* base1, void* base2) {
  MPI_Status status ;

  Mess*    message = messages.data() ;
  MultiBlock block = blocks.data() ;

  for(int mess = 0 ; mess < messages.size() ; mess++) {
    int size    = message->size ;
    int num     = message->num  ;

    char* buffer ;
    if(size > STATIC_BUFFER_SIZE)
          // Only really expected if elementLen 1 + 2 > STATIC_BUFFER_SIZE.
      buffer = new char [size] ;  // *** potential word-alignment problem ***
    else
      buffer = (char*) statBuf ;

    MultiBlock firstBlock = block ;
    int i ;

    // Receive buffer from children, and combine with blocks of array.
    for(i = 0 ; i < numChildren ; i++) {

      MPI_Recv(buffer, size, MPI_CHAR,
	       child [i], DAR, MPI_COMM_WORLD, &status) ;

      MultiBlock chBlock = firstBlock ;
      char* ptr1 = buffer ;
      for(int j = 0 ; j < num ; j++) {
        MultiBlock block1 = chBlock ;
	nextBlock(chBlock) ;
        MultiBlock block2 = chBlock ;
	nextBlock(chBlock) ;
        char* ptr2 = ptr1 + block1 [MBLK_SIZ] ;
	opCombine(block1, block2, ptr1, ptr2) ;
	ptr1 = ptr2 + block2 [MBLK_SIZ] ;
      }
    }

    // Copy blocks of array to buffer
    char* ptr = buffer ;
    for(int j = 0 ; j < num ; j++) {
      gatherMblkBuf(base1, block, ptr, elementLen) ;
      ptr += block [MBLK_SIZ] ;
      nextBlock(block) ;
      gatherMblkBuf(base2, block, ptr, element2Len) ;
      ptr += block [MBLK_SIZ] ;
      nextBlock(block) ;
    }

    if(!isRoot) {
      // Send buffer to parent
      MPI_Send(buffer, size, MPI_CHAR, parent, DAR, MPI_COMM_WORLD) ;

      // Receive buffer from parent
      MPI_Recv(buffer, size, MPI_CHAR, parent, DAR, MPI_COMM_WORLD, &status) ;

      // Copy buffer to blocks of array
      MultiBlock chBlock = firstBlock ;
      char* ptr = buffer ;
      for(int j = 0 ; j < num ; j++) {
	scatterMblkBuf(base1, chBlock, ptr, elementLen) ;
	ptr += chBlock [MBLK_SIZ] ;
	nextBlock(chBlock) ;
	scatterMblkBuf(base2, chBlock, ptr, element2Len) ;
	ptr += chBlock [MBLK_SIZ] ;
	nextBlock(chBlock) ;
      }
    }

    // Send buffer to children
    for(i = 0 ; i < numChildren ; i++)
      MPI_Send(buffer, size, MPI_CHAR, child [i], DAR, MPI_COMM_WORLD) ;

    if(size > STATIC_BUFFER_SIZE)
      delete buffer ;

    message++ ;
  }
}


// ----------------------------------------------------------------------
// Things to do
//
//   Remove `isRoot' from `ProcTree', and add `locId' to these schedules?
//   (or, if it stays, change `isRoot' to `amRoot').
