
// Assorted schedules for 1-sided access, regular remapping and irregular
// subscripting of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
extern "C" {
  #include "armci.h"
}

#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "TreeSchedule.h"
#include "RMSchedule.h"
#include "Reduce.h"
#include "RemapRM.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Get'                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Get :: Get(const int _locStr [], void* locDat,
           const DAD* _rem, const int remOff, void** remPtrs,
           const int len) :
              BlockRMSchedule(locDat, remOff, remPtrs, len, _rem->rnk()),
              locStr(_locStr), rem(_rem), rank(_rem->rnk()) {
  Group q = rem->grp() ;

  exts     = new int [BlockRMSchedule :: rank] ;
  locSteps = new int [BlockRMSchedule :: rank] ;
  remSteps = new int [BlockRMSchedule :: rank] ;

  loop(0, 0, 0, 1, 1, q, 0) ;

  delete [] remSteps ;
  delete [] locSteps ;
  delete [] exts ;
}

void Get :: loop(const int locOffset, const int remOffset,
                 const int numStrLevels, const int contigLen,
                 const int vol,
                 Group remGrp,
                 const int r) {

  if(r == rank)
    transReq(locOffset, locSteps, remOffset, remSteps,
             exts, numStrLevels, contigLen, vol, remGrp.lead_abs()) ;
  else {

    Map    remMap = rem->map(r) ;
    Range  y      = remMap.rng() ;

    const int locStep = locStr [r] ;

    AllBlocksIndex j(y) ;

    for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {

      const int remOffsetElem = remOffset + remMap.offset(j) ;
      const int remStep       = remMap.step(j.sub_stp) ;

      const int jlocStep = locStep * j.glb_stp ;

      if(numStrLevels == 0 && jlocStep == contigLen && remStep == contigLen)
	loop(locOffset + j.glb_bas * locStep, remOffsetElem,
             0, contigLen * j.count,
	     vol * j.count,
	     remGrp / j,
	     r + 1) ;
      else {
	exts     [numStrLevels] = j.count ;
	locSteps [numStrLevels] = jlocStep ;
	remSteps [numStrLevels] = remStep ;

	loop(locOffset + j.glb_bas * locStep, remOffsetElem,
             numStrLevels + 1, contigLen,
	     vol * j.count,
	     remGrp / j,
	     r + 1) ;
      }
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapRMSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RemapRMSkeleton :: RemapRMSkeleton(const DAD* _loc, void* locDat,
                                   const DAD* _rem,
                                   const int remOff, void** remPtrs,
                                   const int len) :
              CollBlockRMSchedule(locDat, remOff, remPtrs,
                                  len, _rem->rnk()),
              loc(_loc), rem(_rem), rank(_loc->rnk()) {
  Group p = loc->grp() ;
  Group q = rem->grp() ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc()) {  // Am part of lead reduced group.
      exts     = new int [BlockRMSchedule :: rank] ;
      locSteps = new int [BlockRMSchedule :: rank] ;
      remSteps = new int [BlockRMSchedule :: rank] ;

      loop(0, 0, 0, 1, 1, q, 0) ;

      delete [] remSteps ;
      delete [] locSteps ;
      delete [] exts ;
    }
  }

  :: spanTree(&spanTree, apg) ;
  CollBlockRMSchedule :: build(&spanTree) ;
}

void RemapRMSkeleton :: loop(const int locOffset, const int remOffset,
                             const int numStrLevels, const int contigLen,
                             const int vol,
                             Group remGrp,
                             const int r) {

  if(r == rank)
    transReq(locOffset, locSteps, remOffset, remSteps,
             exts, numStrLevels, contigLen, vol, remGrp.lead_abs()) ;
  else {

    Range  x      = loc->rng(r) ;
    Map    locMap = loc->map(r) ;

    Range  y      = rem->rng(r) ;
    Map    remMap = rem->map(r) ;

    LocBlocksIndex i(x) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int locOffsetElem = locOffset + locMap.offset(i) ;
      const int locStep = locMap.step(i.sub_stp) ;

      AllBlocksIndex3 j(y, i.count, i.glb_bas, i.glb_stp) ;

      for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {

        const int remOffsetElem = remOffset + remMap.offset(j) ;
        const int remStep       = remMap.step(j.sub_stp) ;

        const int jlocStep = locStep * j.glb_stp ;

	if(numStrLevels == 0 && jlocStep == contigLen && remStep == contigLen)
	  loop(locOffsetElem + j.glb_bas * locStep, remOffsetElem,
	       0, contigLen * j.count,
	       vol * j.count,
	       remGrp / j,
	       r + 1) ;
        else {
	  exts     [numStrLevels] = j.count ;
	  locSteps [numStrLevels] = jlocStep ;
	  remSteps [numStrLevels] = remStep ;

	  loop(locOffsetElem + j.glb_bas * locStep, remOffsetElem,
               numStrLevels + 1, contigLen,
	       vol * j.count,
	       remGrp / j,
	       r + 1) ;
        }
      }
    }
  }
}



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherRMSkeleton'                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

GatherRMSkeleton :: GatherRMSkeleton(const DAD* _loc, void* locDat,
                                     const DAD* _rem,
                                     const int remOff, void** remPtrs,
                                     const int len,
                                     const DAD** _subs, const int** subsDat,
                                     const DAD* _msk, const int* mskDat) :
      CollRMSchedule(locDat, remOff, remPtrs, len),
      loc(_loc), rem(_rem), subs(_subs), msk(_msk), rank(_loc->rnk()) {

  Group p = loc->grp() ;
  q = rem->grp() ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, subsDat, mskDat, 0) ;
  }

  :: spanTree(&spanTree, apg) ;
  CollRMSchedule :: build(&spanTree) ;
}

void GatherRMSkeleton :: loop(int locOffset,
                              const int** subsElem, const int* mskElem,
                              const int r) {

  if(r == rank) {
    if(!msk || *mskElem) {
      Group  remGrp = q ;
      int remOffset = 0 ;

      for(int s = 0 ; s < rem->rnk() ; s++) {
        Range y(rem->rng(s)) ;
        Location i(y(*(subsElem [s]))) ;

        remOffset += rem->map(s).offset(i) ;
        remGrp     = remGrp / i ;
      }

      transReq(locOffset, remGrp.lead_abs(), remOffset) ;
    }
  }
  else {
    LocBlocksIndex i(loc->rng(r)) ;

    Map    locMap = loc->map(r) ;
    Map    mskMap ;
    if(msk) mskMap = msk->map(r) ;

    const int** subsDataElem = new const int* [rem->rnk()] ;
    int* subsStep = new int [rem->rnk()] ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int s ;

      int locOffsetElem = locOffset + locMap.offset(i) ;
      const int locStep = locMap.step(i.sub_stp) ;

      for(s = 0 ; s < rem->rnk() ; s++) {
        Map    subMap = (subs [s])->map(r) ;

        subsDataElem [s] = subsElem [s] + subMap.offset(i) ;
        subsStep [s]     = subMap.step(i.sub_stp) ;
      }

      const int* mskDataElem = msk ? mskElem + mskMap.offset(i) : (int*) NULL ;
      const int mskStep = msk ? mskMap.step(i.sub_stp) : 0 ;

      for(int i_block = 0 ; i_block < i.count ; i_block++) {
        loop(locOffsetElem, subsDataElem, mskDataElem, r + 1) ;

        locOffsetElem += locStep ;
        for(s = 0 ; s < rem->rnk() ; s++)
          subsDataElem [s] += subsStep [s] ;
        if(msk)
          mskDataElem += mskStep ;
      }
    }

    delete [] subsStep ;
    delete [] subsDataElem ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherRMSkeleton'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int VecGatherRMSkeleton :: blockRank(const int rank, const DAD** subs) {
  int result = 0 ;
  for (int s = 0 ; s < rank ; s++)
    if(!(subs [s])) result++ ;

  return result ;
}

VecGatherRMSkeleton :: VecGatherRMSkeleton(const DAD* _loc, void* locDat,
                                           const DAD* _rem,
                                           const int remOff, void** remPtrs,
                                           const int len,
                                           const DAD** _subs,
                                           const int** _subsData) :
              CollBlockRMSchedule(locDat, remOff, remPtrs,
                                  len, blockRank(_rem->rnk(), _subs)),
              loc(_loc), rem(_rem), subs(_subs), subsData(_subsData),
              rank(_loc->rnk()) {
  Group p = loc->grp() ;
  Group q = rem->grp() ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc()) {  // Am part of lead reduced group.
      exts     = new int [BlockRMSchedule :: rank] ;
      locSteps = new int [BlockRMSchedule :: rank] ;
      remSteps = new int [BlockRMSchedule :: rank] ;

      loop(0, 0, 0, 1, 1, q, 0) ;

      delete [] remSteps ;
      delete [] locSteps ;
      delete [] exts ;
    }
  }

  :: spanTree(&spanTree, apg) ;
  CollBlockRMSchedule :: build(&spanTree) ;
}

void VecGatherRMSkeleton :: loop(const int locOffset, const int remOffset,
                                 const int numStrLevels, const int contigLen,
                                 const int vol,
                                 Group remGrp,
                                 const int r) {

  if(r == rank)
    transReq(locOffset, locSteps, remOffset, remSteps,
             exts, numStrLevels, contigLen, vol, remGrp.lead_abs()) ;
  else {

    Range  x      = loc->rng(r) ;
    Map    locMap = loc->map(r) ;

    Range  y      = rem->rng(r) ;
    Map    remMap = rem->map(r) ;

    const DAD* sub = subs [r] ;

    if(sub) {

      // Dimension has a vector subscript.

      Index i(x) ;

      for(i.begin() ; i.test() ; i.next()) {
        const int* subData = subsData [r] ;
        Location j(y(subData [sub->map(0).offset(i)])) ;

        loop(locOffset + locMap.offset(i), remOffset + remMap.offset(j),
             numStrLevels, contigLen,
             vol,
             remGrp / j, r + 1) ;
      }
    }
    else {

      // Dimension is unsubscripted.  (cf `remap').

      LocBlocksIndex i(x) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

        int locOffsetElem = locOffset + locMap.offset(i) ;
        const int locStep = locMap.step(i.sub_stp) ;

        AllBlocksIndex3 j(y, i.count, i.glb_bas, i.glb_stp) ;

        for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {

          const int remOffsetElem = remOffset + remMap.offset(j) ;
          const int remStep       = remMap.step(j.sub_stp) ;

          const int jlocStep = locStep * j.glb_stp ;

	  if(numStrLevels == 0 && jlocStep == contigLen && remStep == contigLen)
	    loop(locOffsetElem + j.glb_bas * locStep, remOffsetElem,
		 0, contigLen * j.count,
		 vol * j.count,
		 remGrp / j,
		 r + 1) ;
          else {
	    exts     [numStrLevels] = j.count ;
	    locSteps [numStrLevels] = jlocStep ;
	    remSteps [numStrLevels] = remStep ;

	    loop(locOffsetElem + j.glb_bas * locStep, remOffsetElem,
                 numStrLevels + 1, contigLen,
		 vol * j.count,
		 remGrp / j,
		 r + 1) ;
          }
        }
      }
    }
  }
}


// -------------------------------------------------------------
// Things to do
//
//   Eliminate use of `Index' iterators in in `VecGatherRMSkeleton',
//       (in favour of lower-level `LocBlocksIndex').
//
//   Where there are several short `int' vectors in a schedule (`exts',
//     `steps', ...) allocate and deallocate them as a single vector.
//
//   Or allocate them using `Vec'?

