//
//     Jacobi iterative method for solving Uxx + Uyy = 1 on [0,1]x[0,1] area
//     Global Arrays are used to store and access U and exact solution vectors 
//     J.N., 06.01.94, works with GA 1.2
//
//     Translated to ad++/ARMCI.  Functionality test only---ad++ performance
//     isn't comparable with Fortran.  DBC 10/5/98.

#include "ad++.h"
#include "adrm.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }

void jacinit(const int m,
	     double& S, double& W, double& C, double& N, double& E,
	     const Section2<double>& D, const Section2<double>& ux) {

    if(m < 1) AdlibError("jacinit: wrong value of m") ;

    S = .25 ;
    W = .25 ;
    N = .25 ;
    E = .25 ;

    const double h = 1.0 / (m + 1) ;

    Range x = ux.rng(0), y = ux.rng(1) ;
    Index i(x), j(y) ;

    OVERALL(j) {
	const int jj = y.idx(j) + 1 ;
	OVERALL(i) {
	    const int ii = x.idx(i) + 1 ;

	    double temp = 0.0 ;
	    if(ii == 1) temp = temp + 1 ;
	    if(jj == 1) temp = temp + 1 ;
	    if(ii == m) temp = temp + jj * h + 1 ;
	    if(jj == m) temp = temp + ii * h + 1 ;
	    D(i,j) = temp / 4.0 ;

	    ux(i, j) = 1.0 + h * h * ii * jj ;
	} ALLOVER(j) ;
    } ALLOVER(i) ;
}

void iterate() {
    const int m = 10 ;
    const double eps = 1E-5 ;           // stopping number
    double S, W, C, N, E ;              // coef. of G matrix and d
    double norm ;

    Procs2 p ;
    ON(p) {
	Dimension dx = p.dim(0), dy = p.dim(1) ;

	BlockRange x(m, dx), y(m, dy) ;

	// *** create global arrays: u - approx. solution,  ux - exact solution
	//                           diff will store u - ux 

	ArrayRM2<double> u(x, y) ;
	Array2<double> ux(x, y), diff(x, y) ;

	// *** check which piece of ux (and u) I own

	Block ib, jb ;
	x.block(&ib, dx.crd()) ;
	y.block(&jb, dy.crd()) ;
	
	const int ni  = ib.count ;
	const int ilo = ib.glb_bas ;
	const int ihi = ilo + ni - 1 ;

	const int nj  = jb.count ;
	const int jlo = jb.glb_bas ;
	const int jhi = jlo + nj - 1 ;

        // *** allocate memory for D

	Array2<double> d(x, y) ;

	// *** create coefficients of the jacobi iteration matrix G and
	//     exact solution

        jacinit(m, S, W, C, N, E, d, ux) ;

        // *** allocate memory for a copy of my piece of u and neighb.
        //     grid points

	BlockRange xg(m, dx, 1, 1), yg(m, dy, 1, 1) ;  // ghost-extended ranges.

	Array2<double> l(xg, yg) ;

        // *** zero the allocated memory

        // [Use low-level inquiries to ensure we reach the ghost regions]

        const int vol = xg.volume() * yg.volume() ;
        for(int ii = 0 ; ii < vol ; ii++)
          l.dat() [ii] = 0.0 ;

        // *** initial guess for u -- zero

        init(u, 0.0) ;

        // *** the stopping test is ||u-ux||/||ux||, ||.|| - second norm

        const double uxnorm = sqrt(dotProduct(ux, ux)) ;          // ||ux||

        // *** Now, synchronize and then iterate 

        int k = 0 ;

        Index i(x), j(y) ;

        loop:
            synch() ;
            k = k + 1 ;

            // ... compute next approximation of u

            OVERALL(i) {
	        OVERALL(j) {
	            u(i, j) = S * l(i + 1, j) + W * l(i, j - 1) +  
		              N * l(i - 1, j) + E * l(i, j + 1) + d(i, j) ;

                    diff(i, j) = u(i, j) - ux(i, j) ;
	        } ALLOVER(j) ;
            } ALLOVER(i) ;

            // ... compute the stopping number -- second norm

            const double product = dotProduct(diff, diff) ;   // <diff, diff>
            norm = sqrt(product) / uxnorm ;
            if(k % 10 == 1) gprintf("%d error= %f\n", k, norm) ;

            if(norm > eps) {
		// ... not done yet, get a copy of my piece of u and
		//     neighboring grid points

		// ... determine where from we should copy data --
		//     consider topology

                // [Use low-level inquiries to find the local patch of `l']

                const int l_str_0 = l.str(0).len(), l_str_1 = l.str(1).len() ;

                // [Adjust base so don't write ghost region if hold low edges]

                double* l_first = l.dat() ;
                if(ilo == 0) l_first += l_str_0 ;
                if(jlo == 0) l_first += l_str_1 ;

                get(l_str_0, l_str_1, l_first,
                    u.sect(Trip(max(0, ilo - 1), min(m - 1, ihi + 1)),
                           Trip(max(0, jlo - 1), min(m - 1, jhi + 1)))) ;

                goto loop ;
            }

        gprintf("converged in %d iterations, error= %f\n", k, norm) ;

        synch() ;
    } NO(p) ;
}

int main(int argc, char* argv[]) {
    AdlibInit(argc, argv) ;
    ARMCI_Init() ;

    iterate() ;                                       // do the work      

    ARMCI_Finalize() ;
    AdlibFinalize() ;
}

